from shutil import copyfile

import pytest

from requests_cache import CachedSession
from tests.conftest import HTTPBIN_FORMATS, SAMPLE_CACHE_FILES, httpbin


@pytest.mark.parametrize('db_path', SAMPLE_CACHE_FILES)
def test_version_upgrade(db_path, tempfile_path):
    """Load SQLite cache files created with older versions of requests-cache.
    Expected behavior: either
    1. Serialization format is incompatible, and previous cache items have been invalidated, or
    2. Serialization format is compatible, and previous cache items are valid

    The previously cached responses should be either replaced or retrieved without any errors.
    """
    copyfile(db_path, tempfile_path)
    session = CachedSession(tempfile_path)

    for response_format in HTTPBIN_FORMATS:
        session.get(httpbin(response_format))
        assert session.get(httpbin(response_format)).from_cache is True



"""Filesystem cache backend. For usage details, see :ref:`Backends: Filesystem <filesystem>`.

.. automodsumm:: requests_cache.backends.filesystem
   :classes-only:
   :nosignatures:
"""

from contextlib import contextmanager
from logging import getLogger
from os import makedirs
from pathlib import Path
from pickle import PickleError
from shutil import rmtree
from threading import RLock
from time import time_ns
from typing import Any, Iterator, Optional

from ..serializers import SerializerType, json_serializer
from . import BaseCache, BaseStorage, StrOrPath
from .sqlite import SQLiteDict, get_cache_path

logger = getLogger(__name__)


class FileCache(BaseCache):
    """Filesystem cache backend.

    Args:
        cache_name: Base directory for cache files
        use_cache_dir: Store database in a user cache directory (e.g., `~/.cache/`)
        use_temp: Store cache files in a temp directory (e.g., ``/tmp/http_cache/``).
            Note: if ``cache_name`` is an absolute path, this option will be ignored.
        decode_content: Decode JSON or text response body into a human-readable format
        extension: Extension for cache files. If not specified, the serializer default extension
            will be used.
        max_cache_bytes: Enable LRU caching, and set the maximum total size (in bytes) of cached
            responses on the file system.
        max_file_bytes: The maximum size of a single file.
            By default, this is the same as ``max_cache_bytes``.
            Only used if ``max_cache_bytes`` is set.
        block_bytes: The size of a block of data on the file system, which will be used when
            computing total file size on disk. Only used if ``max_cache_bytes`` is set.
        sync_index: On startup, sync LRU metadata with any changes on disk since last use. Use this
            if you intend to modify cache files outside of requests-cache. Leave off to reduce
            startup time for larger caches. Only used if ``max_cache_bytes`` is set.
        lock: Replace the default :class:`threading.RLock` object without your own. Use this if you
            want to share the lock between multiple cache instances, and/or use a different lock
            type (such as :py:class:`multiprocessing.RLock` or :py:class:`filelock.FileLock`).
    """

    def __init__(
        self,
        cache_name: StrOrPath = 'http_cache',
        use_temp: bool = False,
        decode_content: bool = True,
        serializer: Optional[SerializerType] = None,
        **kwargs,
    ):
        super().__init__(cache_name=str(cache_name), **kwargs)
        skwargs = {'serializer': serializer, **kwargs} if serializer else kwargs
        self.responses: FileDict = (LRUFileDict if 'max_cache_bytes' in kwargs else FileDict)(
            cache_name, use_temp=use_temp, decode_content=decode_content, **skwargs
        )
        with self.lock:
            self.redirects: SQLiteDict = SQLiteDict(
                self.cache_dir / 'redirects.sqlite', 'redirects', serializer=None, **kwargs
            )

    @property
    def lock(self) -> RLock:
        """The lock used by the cache."""
        return self.responses.lock

    @property
    def cache_dir(self) -> Path:
        """Base directory for cache files"""
        return Path(self.responses.cache_dir)

    def paths(self) -> Iterator[Path]:
        """Get absolute file paths to all cached responses"""
        return self.responses.paths()

    def clear(self):
        """Clear the cache"""
        # FileDict.clear() removes the cache directory, including redirects.sqlite
        self.responses.clear()
        with self.lock:
            self.redirects.init_db()

    def delete(self, *args, **kwargs):
        with self.lock:
            return super().delete(*args, **kwargs)


class FileDict(BaseStorage):
    """A dictionary-like interface to files on the local filesystem.

    The cache directory will be created if it doesn't already exist.
    """

    def __init__(
        self,
        cache_name: StrOrPath,
        use_temp: bool = False,
        use_cache_dir: bool = False,
        extension: Optional[str] = None,
        serializer: Optional[SerializerType] = json_serializer,
        lock: Optional[RLock] = None,
        **kwargs,
    ):
        super().__init__(serializer=serializer, **kwargs)
        self.cache_dir = get_cache_path(cache_name, use_cache_dir=use_cache_dir, use_temp=use_temp)
        self.extension = _get_extension(extension, self.serializer)
        self.is_binary = getattr(self.serializer, 'is_binary', False)
        self._lock = lock if lock is not None else RLock()
        makedirs(self.cache_dir, exist_ok=True)

    @property
    def lock(self) -> RLock:
        """The lock used by the cache."""
        return self._lock

    @contextmanager
    def _try_io(self, key: Optional[str] = None, ignore_errors: bool = False):
        """Attempt an I/O operation, and either ignore errors or re-raise them as KeyErrors"""
        try:
            with self._lock:
                yield
        except (EOFError, IOError, OSError, PickleError) as e:
            if not ignore_errors:
                raise KeyError(f'File for key {key!r} not found.') from e

    def _key2path(self, key: str) -> Path:
        return self.cache_dir / f'{key}{self.extension}'

    def __getitem__(self, key: str):
        mode = 'rb' if self.is_binary else 'r'
        with self._try_io(key):
            with self._key2path(key).open(mode) as f:
                return self.deserialize(key, f.read())

    def __delitem__(self, key):
        with self._try_io(key):
            self._key2path(key).unlink()

    def __setitem__(self, key, value):
        with self._try_io(key):
            with self._key2path(key).open(mode='wb' if self.is_binary else 'w') as f:
                f.write(self.serialize(value))

    def __contains__(self, key) -> bool:
        with self._lock:
            return self._key2path(key).exists()

    def __iter__(self) -> Iterator[str]:
        yield from self.keys()

    def __len__(self) -> int:
        return sum(1 for _ in self.paths())

    def clear(self) -> None:
        """Empty the cache directory."""
        with self._try_io(ignore_errors=True):
            rmtree(self.cache_dir, ignore_errors=True)
            self.cache_dir.mkdir()

    def keys(self):
        return [path.stem for path in self.paths()]

    def paths(self) -> Iterator[Path]:
        """Get absolute file paths to all cached responses"""
        with self._lock:
            return self.cache_dir.glob(f'*{self.extension}')

    def size(self) -> int:
        """Return the size of the database, in bytes"""
        with self._lock:
            return sum(path.stat().st_size for path in self.paths())


class LRUFileDict(FileDict):
    """A size-restricted version of FileDict, using LRU eviction.

    Args:
        block_bytes: The size of a block of data on the file system.
            File sizes will be aligned with this.
        max_cache_bytes: The maximum total size of all files in the cache.
        max_file_bytes: The maximum size of a single file.
            By default, this is the same as ``max_cache_bytes``.
        sync_index: Check for filesystem changes since last use. Use this if you intend to modify
            cache files outside of requests-cache. Leave off to reduce startup time for larger caches.
    """

    def __init__(
        self,
        *args,
        block_bytes: int = 1,
        max_cache_bytes: int = 100 * 1024 * 1024,  # 100MB
        max_file_bytes: Optional[int] = None,
        sync_index: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.max_cache_bytes = max_cache_bytes
        self.block_bytes = block_bytes
        self.max_file_bytes = max_file_bytes or max_cache_bytes
        if self.max_file_bytes > self.max_cache_bytes:
            raise ValueError(
                f'max_file_bytes must be smaller or equal to max_cache_bytes ({max_cache_bytes})'
            )
        if self.block_bytes > self.max_file_bytes:
            raise ValueError(f'block_bytes must be smaller than max_file_bytes ({max_file_bytes})')
        if self.block_bytes < 1:
            raise ValueError(f'block_bytes must be greater than 0, not {block_bytes}')

        self.lru_index = LRUDict(self.cache_dir / 'lru.db', 'lru', **kwargs)
        # Rebuild LRU index if explicitly asked,
        # or for a new cache (potentially with existing files but no metadata)
        if sync_index or len(self.lru_index) == 0:
            self._sync_lru_index()

    def __getitem__(self, key):
        """Get a value and update its access time in the LRU index"""
        value = super().__getitem__(key)

        # Update access time in LRU index
        try:
            self.lru_index.update_access_time(key)
        # File is missing from LRU index
        except KeyError:
            file_path = self._key2path(key)
            if file_path.exists():
                file_size = self._get_size_on_disk(file_path.stat().st_size)
                self.lru_index[key] = file_size

        return value

    def __setitem__(self, key: str, value: Any) -> None:
        content = self.serialize(value)
        data = content.encode('UTF-8') if isinstance(content, str) else content
        del content
        file_size = self._get_size_on_disk(len(data))

        if file_size > self.max_file_bytes:
            logger.debug(
                f'Not caching {key!r} because it is larger than {self.max_file_bytes} bytes.'
            )
            return

        with self._try_io(key):
            try:
                super().__delitem__(key)
            except KeyError:
                pass

            # Make space if needed before writing file
            self._evict(file_size)
            super().__setitem__(key, value)

            # Update LRU index with new file size and access time
            self.lru_index[key] = file_size

    def __delitem__(self, key):
        """Delete a file and remove it from the LRU index"""
        # Remove from LRU index
        try:
            del self.lru_index[key]
        except KeyError:
            pass

        super().__delitem__(key)

    def _evict(self, desired_free_bytes: int):
        """Make space in the cache to fit the given number of bytes, if needed.

        This starts deleting the least recently used entries first.
        """
        current_size = self.lru_index.total_size()
        space_needed = current_size + desired_free_bytes - self.max_cache_bytes
        if space_needed <= 0:
            return

        # Get LRU keys to evict based on how much space we need
        keys_to_evict = self.lru_index.get_lru(space_needed)
        for key in keys_to_evict:
            try:
                del self[key]
            except KeyError:
                pass

    def _sync_lru_index(self):
        """Rebuild the LRU index from files on disk"""
        with self._lock:
            self.lru_index.clear()
            for path in self.paths():
                key = path.stem
                self.lru_index[key] = self._get_size_on_disk(path.stat().st_size)

    def _get_size_on_disk(self, file_size: int) -> int:
        """Return a file size on disk, rounded up to fit the blocks on the file system"""
        sign = -1 if file_size < 0 else 1
        return (
            (file_size * sign + self.block_bytes - 1) // self.block_bytes * self.block_bytes * sign
        )

    def clear(self):
        """Clear the cache directory and LRU index."""
        super().clear()
        self.lru_index.clear()
        self._lru_initialized = False

    def size(self) -> int:
        """Return the size of the database, in bytes"""
        return self.lru_index.total_size()


class LRUDict(SQLiteDict):
    """A SQLite db used to track LRU metadata for cached items:

    * ``key``: The cache key
    * ``access_time``: The last access time, as a UNIX timestamp in nanoseconds
    * ``size``: The size of the cached item, in bytes
    * ``total_size`` Combined size of all cache items in bytes, accessed with :py:meth:`total_size`.

    Implementation Notes:

    * ``total_size`` is managed by triggers and stored in a separate single-row table.
    * As a dict-like interface, ``size`` is treated as the main value and ``access_time`` is set
      automatically and updated with :py:meth:`update_access_time`.
    * :py:meth:`get_lru` Can select multiple keys to evict, up to an arbitrary total size, within a
      single query using a window function.
    """

    def __init__(self, *args, **kwargs):
        kwargs.pop('serializer', None)
        super().__init__(*args, **kwargs)

    def init_db(self):
        self.close()
        with self.connection(commit=True) as con:
            # Table for LRU metadata
            con.execute(
                f'CREATE TABLE IF NOT EXISTS {self.table_name} ('
                '    key TEXT PRIMARY KEY,'
                '    access_time INTEGER NOT NULL,'
                '    size INTEGER NOT NULL'
                ')'
            )
            con.execute(
                f'CREATE INDEX IF NOT EXISTS idx_access_time ON {self.table_name}(access_time)'
            )
            con.execute(f'CREATE INDEX IF NOT EXISTS idx_size ON {self.table_name}(size)')

            # Single-row table to persist total cache size
            con.execute(
                f'CREATE TABLE IF NOT EXISTS {self.table_name}_size ('
                '    total_size INTEGER NOT NULL'
                ')'
            )
            con.execute(f'INSERT OR IGNORE INTO {self.table_name}_size (total_size) VALUES (0)')

            # Triggers to update total size
            con.execute(
                f"""
                CREATE TRIGGER IF NOT EXISTS {self.table_name}_insert
                AFTER INSERT ON {self.table_name}
                BEGIN
                    UPDATE {self.table_name}_size
                    SET total_size = total_size + NEW.size;
                END;
                """
            )
            con.execute(
                f"""
                CREATE TRIGGER IF NOT EXISTS {self.table_name}_delete
                AFTER DELETE ON {self.table_name}
                BEGIN
                    UPDATE {self.table_name}_size
                    SET total_size = total_size - OLD.size;
                END;
                """
            )
            con.execute(
                f"""
                CREATE TRIGGER IF NOT EXISTS {self.table_name}_update
                AFTER UPDATE OF size ON {self.table_name}
                WHEN OLD.size != NEW.size
                BEGIN
                    UPDATE {self.table_name}_size
                    SET total_size = total_size + (NEW.size - OLD.size);
                END;
                """
            )

    def __delitem__(self, key):
        with self.connection(commit=True) as con:
            cur = con.execute(f'DELETE FROM {self.table_name} WHERE key=?', (key,))
        if not cur.rowcount:
            raise KeyError

    def __getitem__(self, key) -> int:
        with self.connection() as con:
            # Using placeholders here with python 3.12+ and concurrency results in the error:
            # sqlite3.InterfaceError: bad parameter or other API misuse
            row = con.execute(f"SELECT size FROM {self.table_name} WHERE key='{key}'").fetchone()
            if not row:
                raise KeyError(key)
            return row[0]

    def __setitem__(self, key: str, size: int):
        """Save a value (file size), and update access time and total cache size"""

        timestamp = int(time_ns())
        with self.connection(commit=True) as con:
            con.execute(
                f"""
                INSERT INTO {self.table_name} (key, access_time, size)
                VALUES (?, ?, ?)
                ON CONFLICT(key) DO UPDATE
                SET access_time = excluded.access_time, size = excluded.size
                """,
                (key, timestamp, size),
            )

    def clear(self):
        super().clear()
        with self.connection(commit=True) as con:
            con.execute(f'UPDATE {self.table_name}_size SET total_size = 0')

    def count(self, *args, **kwargs):
        with self.connection() as con:
            return con.execute(f'SELECT COUNT(key) FROM {self.table_name}').fetchone()[0]

    def get_lru(self, total_size: int):
        """Get the least recently used keys with a combined size >= total_size"""

        with self.connection() as con:
            cur = con.execute(
                f"""
                WITH ordered AS (
                    SELECT key, size, access_time, SUM(size) OVER (ORDER BY access_time) AS running_total
                    FROM {self.table_name}
                )
                SELECT * FROM ordered WHERE running_total - size < ?
                ORDER BY access_time;
                """,
                (total_size,),
            )
            rows = cur.fetchall()
            cur.close()
            return [row[0] for row in rows]

    def sorted(  # type: ignore
        self,
        key: str = 'access_time',
        reversed: bool = False,
        limit: Optional[int] = None,
        **kwargs,
    ) -> Iterator[str]:
        """Get LRU entries in sorted order, by either ``access_time`` or ``size``"""
        # Get sort key, direction, and limit
        if key not in ['access_time', 'size', 'key']:
            raise ValueError(f'Invalid sort key: {key}')
        direction = 'DESC' if reversed else 'ASC'
        limit_expr = f'LIMIT {limit}' if limit else ''

        with self.connection() as con:
            for row in con.execute(
                f'SELECT key FROM {self.table_name} ORDER BY {key} {direction} {limit_expr}',
            ):
                yield row[0]

    def total_size(self) -> int:
        with self.connection() as con:
            row = con.execute(f'SELECT total_size FROM {self.table_name}_size').fetchone()
            return row[0] if row else 0

    def update_access_time(self, key: str):
        """Update the given key with the current timestamp

        Raises:
            KeyError: If the key doesn't exist in the LRU index
        """
        timestamp = int(time_ns())
        with self.connection(commit=True) as con:
            cur = con.execute(
                f'UPDATE {self.table_name} SET access_time = ? WHERE key = ?',
                (timestamp, key),
            )
        if not cur.rowcount:
            raise KeyError(key)


def _get_extension(extension: Optional[str] = None, serializer=None) -> str:
    """Use either the provided file extension, or get the serializer's default extension"""
    if extension:
        return f'.{extension}'
    subs = {
        'bson': 'bson',
        'safe_pickle': 'pkl',
        'pickle': 'pkl',
        'orjson': 'json',
        'ujson': 'json',
    }
    if serializer and (name := serializer.name):
        for k, v in subs.items():
            name = name.replace(k, v)
        return f'.{name}'
    return '.dat'



"""BaseCache tests that use mocked responses only"""

import pickle
from datetime import timedelta
from logging import getLogger
from pickle import PickleError
from unittest.mock import patch

import pytest
from requests import Request

from requests_cache.backends import BaseCache, SQLiteDict
from requests_cache.cache_keys import create_key
from requests_cache.models import CachedRequest, CachedResponse
from requests_cache.policy import utcnow
from requests_cache.session import CachedSession
from tests.conftest import (
    MOCKED_URL,
    MOCKED_URL_ETAG,
    MOCKED_URL_HTTPS,
    MOCKED_URL_JSON,
    MOCKED_URL_REDIRECT,
    START_DT,
    ignore_deprecation,
    mount_mock_adapter,
    patch_normalize_url,
    skip_pypy,
    time_travel,
)

logger = getLogger(__name__)


class InvalidResponse:
    """Class that will raise an error when unpickled"""

    def __init__(self):
        self.foo = 'bar'

    def __setstate__(self, value):
        raise ValueError('Invalid response!')


def test_contains__key(mock_session):
    mock_session.get(MOCKED_URL, params={'foo': 'bar'})
    key = list(mock_session.cache.responses.keys())[0]
    assert mock_session.cache.contains(key)
    assert not mock_session.cache.contains(f'{key}_b')


def test_contains__request(mock_session):
    mock_session.get(MOCKED_URL, params={'foo': 'bar'})
    request = Request('GET', MOCKED_URL, params={'foo': 'bar'})
    assert mock_session.cache.contains(request=request)
    request.params = None
    assert not mock_session.cache.contains(request=request)


def test_contains__multipart_request(mock_session):
    files = [('file', ('test.txt', b'content'))]
    mock_session.settings.allowable_methods = ('GET', 'POST')
    mock_session.post(MOCKED_URL, files=files)
    request = Request('POST', MOCKED_URL, files=files)
    assert mock_session.cache.contains(request=request)
    request.files = None
    assert not mock_session.cache.contains(request=request)


@pytest.mark.parametrize('verify', [True, False])
def test_contains__url(verify, mock_session):
    mock_session.get(MOCKED_URL, verify=verify)
    assert mock_session.cache.contains(url=MOCKED_URL, verify=verify)
    assert not mock_session.cache.contains(url=f'{MOCKED_URL}?foo=bar', verify=verify)


def test_delete__keys(mock_session):
    r = mock_session.get(MOCKED_URL)
    mock_session.cache.delete(r.cache_key, 'nonexistent_key')
    mock_session.cache.delete('nonexistent_key')
    assert not mock_session.cache.contains(url=MOCKED_URL)


@skip_pypy  # time-machine doesn't work on PyPy
@patch_normalize_url
def test_delete__expired(mock_normalize_url, mock_session):
    unexpired_url = f'{MOCKED_URL}?x=1'
    mock_session.mock_adapter.register_uri(
        'GET', unexpired_url, status_code=200, text='mock response'
    )
    mock_session.settings.expire_after = 1

    with time_travel(START_DT):
        mock_session.get(MOCKED_URL)
        mock_session.get(MOCKED_URL_JSON)

    mock_session.settings.expire_after = 2
    with time_travel(START_DT + timedelta(seconds=1.1)):
        mock_session.get(unexpired_url)

    # At this point we should have 1 unexpired response and 2 expired responses
    assert len(mock_session.cache.responses) == 3

    # Use the generic BaseCache implementation, not the SQLite-specific one
    with time_travel(START_DT + timedelta(seconds=2)):
        BaseCache.delete(mock_session.cache, expired=True)
    assert len(mock_session.cache.responses) == 1

    cached_response = list(mock_session.cache.responses.values())[0]
    assert cached_response.url == unexpired_url

    # Now the last response should be expired as well
    with time_travel(START_DT + timedelta(seconds=4)):
        BaseCache.delete(mock_session.cache, expired=True)
    assert len(mock_session.cache.responses) == 0


@skip_pypy
def test_delete__expired__per_request(mock_session):
    second_url = f'{MOCKED_URL}/endpoint_2'
    third_url = f'{MOCKED_URL}/endpoint_3'
    mock_session.mock_adapter.register_uri('GET', second_url, status_code=200)
    mock_session.mock_adapter.register_uri('GET', third_url, status_code=200)

    # Cache 3 responses with different expiration times
    with time_travel(START_DT):
        mock_session.get(MOCKED_URL)
        mock_session.get(second_url, expire_after=5)
        mock_session.get(third_url, expire_after=10)

        # All 3 responses should still be cached
        mock_session.cache.delete(expired=True)
        for response in mock_session.cache.responses.values():
            logger.info(f'Expires in {response.expires_delta} seconds')
        assert len(mock_session.cache.responses) == 3

    # One should be expired after 5s, and another should be expired after 10s
    with time_travel(START_DT + timedelta(seconds=6)):
        mock_session.cache.delete(expired=True)
        assert len(mock_session.cache.responses) == 2
    with time_travel(START_DT + timedelta(seconds=11)):
        mock_session.cache.delete(expired=True)
        assert len(mock_session.cache.responses) == 1


def test_delete__invalid(tempfile_path):
    class BadSerialzier:
        def dumps(self, value):
            return pickle.dumps(value)

        def loads(self, value):
            response = pickle.loads(value)
            if response.url.endswith('/json'):
                raise PickleError
            return response

    mock_session = CachedSession(
        cache_name=tempfile_path, backend='sqlite', serializer=BadSerialzier()
    )
    mock_session = mount_mock_adapter(mock_session)

    # Start with two cached responses, one of which will raise an error
    mock_session.get(MOCKED_URL)
    mock_session.get(MOCKED_URL_JSON)

    # Use the generic BaseCache implementation, not the SQLite-specific one
    BaseCache.delete(mock_session.cache, expired=True, invalid=True)

    assert len(mock_session.cache.responses) == 1
    assert mock_session.get(MOCKED_URL).from_cache is True
    assert mock_session.get(MOCKED_URL_JSON).from_cache is False


@skip_pypy
def test_delete__older_than(mock_session):
    # Cache 4 responses with different creation times
    with time_travel(START_DT):
        request = CachedRequest(method='GET', url='https://test.com/test_0')
        mock_session.cache.save_response(CachedResponse(request=request))
    with time_travel(START_DT - timedelta(seconds=1.1)):
        request = CachedRequest(method='GET', url='https://test.com/test_1')
        mock_session.cache.save_response(CachedResponse(request=request))
    with time_travel(START_DT - timedelta(seconds=2.1)):
        request = CachedRequest(method='GET', url='https://test.com/test_2')
        mock_session.cache.save_response(CachedResponse(request=request))
    with time_travel(START_DT - timedelta(seconds=3.1)):
        request = CachedRequest(method='GET', url='https://test.com/test_3')
        mock_session.cache.save_response(CachedResponse(request=request))

    # Incrementally remove responses older than 3, 2, and 1 seconds
    with time_travel(START_DT):
        assert len(mock_session.cache.responses) == 4
        mock_session.cache.delete(older_than=timedelta(seconds=3))
        assert len(mock_session.cache.responses) == 3
        mock_session.cache.delete(older_than=timedelta(seconds=2))
        assert len(mock_session.cache.responses) == 2
        mock_session.cache.delete(older_than=timedelta(seconds=1))
        assert len(mock_session.cache.responses) == 1

    # Remove the last response after it's 1 second old
    with time_travel(START_DT + timedelta(seconds=1.1)):
        mock_session.cache.delete(older_than=timedelta(seconds=1))
    assert len(mock_session.cache.responses) == 0


def test_delete__urls(mock_session):
    urls = [MOCKED_URL, MOCKED_URL_JSON, MOCKED_URL_REDIRECT]
    for url in urls:
        mock_session.get(url)

    mock_session.cache.delete(urls=urls)

    for url in urls:
        assert not mock_session.cache.contains(url=url)


@pytest.mark.parametrize('verify', [True, False])
def test_delete__requests(verify, mock_session):
    urls = [MOCKED_URL, MOCKED_URL_JSON, MOCKED_URL_REDIRECT]
    for url in urls:
        mock_session.get(url, verify=verify)

    requests = [Request('GET', url).prepare() for url in urls]
    mock_session.cache.delete(requests=requests, verify=verify)

    for request in requests:
        assert not mock_session.cache.contains(request=request)


def test_recreate_keys(mock_session):
    # Cache some initial responses with default key function
    urls = [MOCKED_URL, MOCKED_URL_JSON, MOCKED_URL_ETAG]
    for url in urls:
        mock_session.get(url)
    old_cache_keys = set(mock_session.cache.responses.keys())

    # Switch to a new key function and recreate keys
    def new_key_fn(*args, **kwargs):
        return create_key(*args, **kwargs) + '_suffix'

    # Check that responses are saved with new keys
    mock_session.settings.key_fn = new_key_fn
    mock_session.cache.recreate_keys()
    new_cache_keys = set(mock_session.cache.responses.keys())
    assert len(old_cache_keys) == len(new_cache_keys) == len(urls)
    assert old_cache_keys != new_cache_keys

    # Check that responses are returned from the cache correctly using the new key function
    for url in urls:
        assert mock_session.get(url).from_cache is True


def test_recreate_keys__same_key_fn(mock_session):
    urls = [MOCKED_URL, MOCKED_URL_JSON, MOCKED_URL_ETAG]
    for url in urls:
        mock_session.get(url)
    old_cache_keys = set(mock_session.cache.responses.keys())

    mock_session.cache.recreate_keys()
    new_cache_keys = set(mock_session.cache.responses.keys())
    assert old_cache_keys == new_cache_keys

    # Check that responses are returned from the cache correctly using the new key function
    for url in urls:
        assert mock_session.get(url).from_cache is True


def test_recreate_keys__empty_response_body(mock_session):
    # Cache some initial responses with default key function, and modify request body (pre-1.0)
    urls = [MOCKED_URL, MOCKED_URL_JSON, MOCKED_URL_ETAG]
    for url in urls:
        r = mock_session.get(url)
        r = CachedResponse.from_response(r)
        r.request.body = b'None'
        mock_session.cache.responses[r.cache_key] = r

    # Switch to a new key function and recreate keys
    def new_key_fn(*args, **kwargs):
        return create_key(*args, **kwargs) + '_suffix'

    mock_session.settings.key_fn = new_key_fn
    mock_session.cache.recreate_keys()

    # Check that responses are returned from the cache correctly using the new key function
    for url in urls:
        assert mock_session.get(url).from_cache is True


def test_reset_expiration__extend_expiration(mock_session):
    # Start with an expired response
    mock_session.settings.expire_after = utcnow() - timedelta(seconds=1)
    mock_session.get(MOCKED_URL)

    # Set expiration in the future
    mock_session.cache.reset_expiration(utcnow() + timedelta(seconds=1))
    assert len(mock_session.cache.responses) == 1
    response = mock_session.get(MOCKED_URL)
    assert response.is_expired is False and response.from_cache is True


def test_reset_expiration__shorten_expiration(mock_session):
    # Start with a non-expired response
    mock_session.settings.expire_after = utcnow() + timedelta(seconds=1)
    mock_session.get(MOCKED_URL)

    # Set expiration in the past
    mock_session.cache.reset_expiration(utcnow() - timedelta(seconds=1))
    response = mock_session.get(MOCKED_URL)
    assert response.is_expired is False and response.from_cache is False


def test_clear(mock_session):
    mock_session.get(MOCKED_URL)
    mock_session.get(MOCKED_URL_REDIRECT)
    mock_session.cache.clear()
    assert not mock_session.cache.contains(url=MOCKED_URL)
    assert not mock_session.cache.contains(url=MOCKED_URL_REDIRECT)


def test_save_response__manual(mock_session):
    response = mock_session.get(MOCKED_URL)
    mock_session.cache.clear()
    mock_session.cache.save_response(response)


def test_update(mock_session):
    src_cache = BaseCache()
    for i in range(20):
        src_cache.responses[f'key_{i}'] = f'value_{i}'
        src_cache.redirects[f'key_{i}'] = f'value_{i}'

    mock_session.cache.update(src_cache)
    assert len(mock_session.cache.responses) == 20
    assert len(mock_session.cache.redirects) == 20


@patch_normalize_url
def test_urls(mock_normalize_url, mock_session):
    for url in [MOCKED_URL, MOCKED_URL, MOCKED_URL_JSON, MOCKED_URL_HTTPS]:
        mock_session.get(url)

    expected_urls = [MOCKED_URL_JSON, MOCKED_URL, MOCKED_URL_HTTPS]
    assert mock_session.cache.urls() == expected_urls


def test_urls__error(mock_session):
    responses = [mock_session.get(url) for url in [MOCKED_URL, MOCKED_URL_JSON, MOCKED_URL_HTTPS]]
    responses[2] = None
    with patch.object(SQLiteDict, 'deserialize', side_effect=responses):
        expected_urls = [MOCKED_URL_JSON, MOCKED_URL]
        assert mock_session.cache.urls() == expected_urls

    # The invalid response should be skipped, but remain in the cache
    assert len(mock_session.cache.responses.keys()) == 3


# Deprecated methods
# --------------------


def test_remove_expired_responses(mock_session):
    with ignore_deprecation(), patch.object(
        mock_session.cache, 'delete'
    ) as mock_delete, patch.object(mock_session.cache, 'reset_expiration') as mock_reset:
        mock_session.cache.remove_expired_responses(expire_after=1)
        mock_delete.assert_called_once_with(expired=True, invalid=True)
        mock_reset.assert_called_once_with(1)

        mock_session.cache.remove_expired_responses()
        assert mock_delete.call_count == 2 and mock_reset.call_count == 1



#!/usr/bin/env python
# fmt: off
"""
An example of {ref}`url-patterns`
"""
from datetime import timedelta

from requests_cache import DO_NOT_CACHE, NEVER_EXPIRE, CachedSession

default_expire_after = 60 * 60               # By default, cached responses expire in an hour
urls_expire_after = {
    'httpbin.org/image': timedelta(days=7),  # Requests for this base URL will expire in a week
    '*.fillmurray.com': NEVER_EXPIRE,        # Requests matching this pattern will never expire
    '*.placeholder.com/*': DO_NOT_CACHE,     # Requests matching this pattern will not be cached
}
urls = [
    'https://httpbin.org/get',               # Will expire in an hour
    'https://httpbin.org/image/jpeg',        # Will expire in a week
    'https://www.fillmurray.com/460/300',    # Will never expire
    'https://via.placeholder.com/350x150',   # Will not be cached
]


def send_requests():
    session = CachedSession(
        cache_name='example_cache',
        expire_after=default_expire_after,
        urls_expire_after=urls_expire_after,
    )
    return [session.get(url) for url in urls]


def _expires_str(response):
    if not response.from_cache:
        return 'N/A'
    elif response.expires is None:
        return 'Never'
    else:
        return response.expires.isoformat()


def main():
    send_requests()
    cached_responses = send_requests()
    for response in cached_responses:
        print(
            f'{response.url:40} From cache: {response.from_cache:}'
            f'\tExpires: {_expires_str(response)}'
        )


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# flake8: noqa: F841
"""
The same as `basic_sessions.py`, but using {ref}`patching`
"""
import time

import requests

import requests_cache

# After installation, all requests functions and Session methods will be cached
requests_cache.install_cache('example_cache', backend='sqlite')


def main():
    # The real request will only be made once; afterward, the cached response is used
    for _ in range(5):
        response = requests.get('https://httpbin.org/get')

    # This is more obvious when calling a slow endpoint
    for _ in range(5):
        response = requests.get('https://httpbin.org/delay/2')

    # Caching can be disabled if we want to get a fresh page and not cache it
    with requests_cache.disabled():
        print(requests.get('https://httpbin.org/ip').text)

    # Get some debugging info about the cache
    print(requests_cache.get_cache())
    print('Cached URLS:')
    print('\n'.join(requests_cache.get_cache().urls()))

    # Uninstall to remove caching from all requests functions
    requests_cache.uninstall_cache()


if __name__ == '__main__':
    t = time.time()
    main()
    print('Elapsed: %.3f seconds' % (time.time() - t))



"""Common tests to run for all backends (BaseStorage subclasses)"""

from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from typing import Dict, Type

import pytest
from attrs import define, field

from requests_cache.backends import BaseStorage
from requests_cache.models import CachedResponse
from tests.conftest import CACHE_NAME, N_ITERATIONS, N_REQUESTS_PER_ITERATION, N_WORKERS


class BaseStorageTest:
    """Base class for testing cache storage dict-like interfaces"""

    storage_class: Type[BaseStorage] = None
    init_kwargs: Dict = {}
    num_instances: int = 10  # Max number of cache instances to test

    def init_cache(self, cache_name=CACHE_NAME, index=0, clear=True, **kwargs):
        kwargs = {**self.init_kwargs, **kwargs}
        cache = self.storage_class(cache_name, f'table_{index}', **kwargs)
        if clear:
            cache.clear()
        return cache

    def teardown_class(cls):
        for i in range(cls.num_instances):
            cls().init_cache(index=i, clear=True)

    def test_basic_methods(self):
        """Test basic dict methods with multiple cache instances:
        ``getitem, setitem, delitem, len, contains``
        """
        caches = [self.init_cache(index=i) for i in range(10)]
        for i in range(self.num_instances):
            caches[i][f'key_{i}'] = f'value_{i}'
            caches[i][f'key_{i + 1}'] = f'value_{i + 1}'

        for i in range(self.num_instances):
            cache = caches[i]
            assert cache[f'key_{i}'] == f'value_{i}'
            assert len(cache) == 2
            assert f'key_{i}' in cache and f'key_{i + 1}' in cache

            del cache[f'key_{i}']
            assert f'key_{i}' not in cache

    def test_iterable_methods(self):
        """Test iterable dict methods with multiple cache instances:
        ``iter, keys, values, items``
        """
        caches = [self.init_cache(index=i) for i in range(self.num_instances)]
        for i in range(self.num_instances):
            caches[i][f'key_{i}'] = f'value_{i}'

        for i in range(self.num_instances):
            cache = caches[i]
            assert list(cache) == [f'key_{i}']
            assert list(cache.keys()) == [f'key_{i}']
            assert list(cache.values()) == [f'value_{i}']
            assert list(cache.items()) == [(f'key_{i}', f'value_{i}')]
            assert dict(cache) == {f'key_{i}': f'value_{i}'}

    def test_cache_key(self):
        """The cache_key attribute should be available on responses returned from all
        mapping/collection methods
        """
        cache = self.init_cache()
        cache['key'] = CachedResponse()
        assert cache['key'].cache_key == 'key'
        assert list(cache.values())[0].cache_key == 'key'
        assert list(cache.items())[0][1].cache_key == 'key'

    def test_del(self):
        """Some more tests to ensure ``delitem`` deletes only the expected items"""
        cache = self.init_cache()
        for i in range(20):
            cache[f'key_{i}'] = f'value_{i}'
        for i in range(5):
            del cache[f'key_{i}']

        assert len(cache) == 15
        assert set(cache.keys()) == {f'key_{i}' for i in range(5, 20)}
        assert set(cache.values()) == {f'value_{i}' for i in range(5, 20)}

    def test_bulk_delete(self):
        cache = self.init_cache()
        for i in range(20):
            cache[f'key_{i}'] = f'value_{i}'
        cache.bulk_delete([f'key_{i}' for i in range(5)])
        cache.bulk_delete(['nonexistent_key'])

        assert len(cache) == 15
        assert set(cache.keys()) == {f'key_{i}' for i in range(5, 20)}
        assert set(cache.values()) == {f'value_{i}' for i in range(5, 20)}

    def test_bulk_delete__noop(self):
        """Just make sure bulk_delete doesn't do anything unexpected if no keys are provided"""
        cache = self.init_cache()
        for i in range(20):
            cache[f'key_{i}'] = f'value_{i}'
        cache.bulk_delete([])
        assert len(cache) == 20

    def test_keyerrors(self):
        """Accessing or deleting a deleted item should raise a KeyError"""
        cache = self.init_cache()
        cache['key'] = 'value'
        del cache['key']

        with pytest.raises(KeyError):
            del cache['key']
        with pytest.raises(KeyError):
            cache['key']

    def test_deleting_key_that_does_not_exist_does_not_delete_other_keys(self):
        """If a key does not exist, deleting it should not delete other keys."""
        cache = self.init_cache()
        cache['key'] = 'value'
        with pytest.raises(KeyError):
            del cache['absent']
        assert len(cache) == 1
        assert 'key' in cache

    def test_picklable_dict(self):
        cache = self.init_cache(serializer='pickle')
        original_obj = BasicDataclass(
            bool_attr=True,
            datetime_attr=datetime(2022, 2, 2),
            int_attr=2,
            str_attr='value',
        )
        cache['key_1'] = original_obj

        obj = cache['key_1']
        assert obj.bool_attr == original_obj.bool_attr
        assert obj.datetime_attr == original_obj.datetime_attr
        assert obj.int_attr == original_obj.int_attr
        assert obj.str_attr == original_obj.str_attr

    def test_clear(self):
        cache_1 = self.init_cache()
        cache_2 = self.init_cache(connection=getattr(cache_1, 'connection', None))

        for i in range(5):
            cache_1[f'key_{i}'] = f'value_{i}'
            cache_2[f'key_{i}'] = f'value_{i}'

        assert len(cache_1) == len(cache_2) == 5
        cache_1.clear()
        cache_2.clear()
        assert len(cache_1) == len(cache_2) == 0
        assert not cache_1 and not cache_2
        assert list(cache_1.keys()) == list(cache_2.keys()) == []

    def test_clear_empty_cache(self):
        cache = self.init_cache()
        cache.clear()
        assert len(cache) == 0
        assert list(cache.keys()) == []

    def test_same_settings(self):
        cache_1 = self.init_cache()
        cache_2 = self.init_cache(connection=getattr(cache_1, 'connection', None))
        cache_1['key_1'] = 'value_1'
        cache_2['key_2'] = 'value_2'
        assert cache_1 == cache_2

    def test_str(self):
        """Not much to test for __str__ methods, just make sure they return keys in some format"""
        cache = self.init_cache()
        for i in range(10):
            cache[f'key_{i}'] = f'value_{i}'
        for i in range(10):
            assert f'key_{i}' in str(cache)

    def test_concurrency(self):
        """Test a large number of concurrent write operations for each backend"""
        cache = self.init_cache()

        def write(i):
            cache[f'key_{i}'] = f'value_{i}'

        n_iterations = N_ITERATIONS * N_REQUESTS_PER_ITERATION * 10
        with ThreadPoolExecutor(max_workers=N_WORKERS * 2) as executor:
            _ = list(executor.map(write, range(n_iterations)))

    def test_bool(self):
        """Check boolean conversion."""
        cache = self.init_cache()
        assert not cache
        cache['key'] = 'value'
        assert cache
        del cache['key']
        assert not cache

    def test_reused_keys_can_be_removed(self):
        """Keys can be reused and can be removed."""
        cache = self.init_cache()
        cache['key'] = 'value'
        del cache['key']
        assert list(cache.keys()) == []
        cache['key'] = 'value'
        cache['key2'] = 'value2'
        del cache['key']
        assert list(cache.keys()) == ['key2']
        del cache['key2']
        assert list(cache.keys()) == []

    def test_values_are_replaced(self):
        """Check that keys are added."""
        cache = self.init_cache()
        cache['key'] = 'value'
        assert list(cache.keys()) == ['key']
        cache['key'] = 'value2'
        assert cache['key'] == 'value2'
        assert list(cache.keys()) == ['key']
        cache['key2'] = 'value3'
        assert set(cache.keys()) == {'key', 'key2'}
        assert cache['key'] == 'value2'
        assert cache['key2'] == 'value3'


@define
class BasicDataclass:
    bool_attr: bool = field(default=None)
    datetime_attr: datetime = field(default=None)
    int_attr: int = field(default=None)
    str_attr: str = field(default=None)



#!/usr/bin/env python
"""
Example of a custom {ref}`request matcher <matching>` that caches a new response if the version of
requests-cache, requests, or urllib3 changes.

This generally isn't needed, since anything that causes a deserialization error will simply result
in a new request being sent and cached. But you might want to include a library version in your cache
key if, for example, you suspect a change in the library does not cause errors but **results in
different response content**.

This uses info from {py:func}`requests.help.info`. You can also preview this info from the command
line to see what else is available:
```bash
python -m requests.help
```
"""
from hashlib import sha256
from importlib.metadata import version as pkg_version
from unittest.mock import patch

from requests import PreparedRequest
from requests.help import info as get_requests_info

from requests_cache import CachedSession
from requests_cache.cache_keys import create_key


def create_custom_key(request: PreparedRequest, **kwargs) -> str:
    """Make a custom cache key that includes library versions"""
    # Start with the default key created by requests-cache
    base_key = create_key(request, **kwargs)
    key = sha256()
    key.update(base_key.encode('utf-8'))

    # Add versions of requests-cache, requests, and urllib3
    requests_info = get_requests_info()
    for lib in ['requests', 'urllib3']:
        key.update(requests_info[lib]['version'].encode('utf-8'))
    key.update(pkg_version('requests_cache').encode('utf-8'))

    return key.hexdigest()


def test_cache_key():
    """Test that the custom cache keys are working as expected"""
    session = CachedSession('key-test', key_fn=create_custom_key)
    session.cache.clear()
    session.get('https://httpbin.org/get')
    response = session.get('https://httpbin.org/get')
    assert response.from_cache is True

    # Simulate a major version change
    new_versions = {
        'requests': {'version': '3.0.0'},
        'urllib3': {'version': '2.0.0'},
    }
    with patch('__main__.get_requests_info', return_value=new_versions):
        # A new request will be sent since the cache key no longer matches
        response = session.get('https://httpbin.org/get')
        assert response.from_cache is False


if __name__ == '__main__':
    test_cache_key()



"""Example of using requests-cache with the requests-mock library"""

import pytest
import requests

import requests_cache


@pytest.fixture(scope='function')
def requests_cache_mock(requests_mock):
    with requests_cache.disabled():
        yield requests_mock


def test_requests_cache_mock(requests_cache_mock):
    """Within this test function, requests will be mocked and not cached"""
    url = 'https://example.com'
    requests_cache_mock.get(url, text='Mock response!')

    # Make sure the mocker is used
    response_1 = requests.get(url)
    assert response_1.text == 'Mock response!'

    # Make sure the cache is not used
    response_2 = requests.get(url)
    assert getattr(response_2, 'from_cache', False) is False



"""Classes and functions for cache persistence. See :ref:`backends` for general usage info."""

# ruff: noqa: F401
from logging import getLogger
from pathlib import Path
from typing import Callable, Dict, Iterable, Optional, Type, Union
from warnings import warn

from .._utils import get_placeholder_class, get_valid_kwargs
from .base import BaseCache, BaseStorage, DictStorage

# Backend-specific keyword arguments equivalent to 'cache_name'
CACHE_NAME_KWARGS = ['db_path', 'db_name', 'namespace', 'table_name']

BackendSpecifier = Union[str, BaseCache]
StrOrPath = Union[Path, str]
logger = getLogger(__name__)


# Import all backend classes for which dependencies are installed
try:
    from .dynamodb import DynamoDbCache, DynamoDbDict
except ImportError as e:
    DynamoDbCache = DynamoDbDict = get_placeholder_class(e)  # type: ignore

try:
    from .gridfs import GridFSCache, GridFSDict
except ImportError as e:
    GridFSCache = GridFSDict = get_placeholder_class(e)  # type: ignore

try:
    from .mongodb import MongoCache, MongoDict
except ImportError as e:
    MongoCache = MongoDict = get_placeholder_class(e)  # type: ignore

try:
    from .redis import RedisCache, RedisDict, RedisHashDict
except ImportError as e:
    RedisCache = RedisDict = RedisHashDict = get_placeholder_class(e)  # type: ignore

try:
    from .sqlite import SQLiteCache, SQLiteDict
except ImportError as e:
    SQLiteCache = SQLiteDict = get_placeholder_class(e)  # type: ignore

try:
    from .filesystem import FileCache, FileDict, LRUFileDict
except ImportError as e:
    FileCache = FileDict = LRUFileDict = get_placeholder_class(e)  # type: ignore


BACKEND_CLASSES = {
    'dynamodb': DynamoDbCache,
    'filesystem': FileCache,
    'gridfs': GridFSCache,
    'memory': BaseCache,
    'mongodb': MongoCache,
    'redis': RedisCache,
    'sqlite': SQLiteCache,
}


def init_backend(
    cache_name: StrOrPath, backend: Optional[BackendSpecifier] = None, **kwargs
) -> BaseCache:
    """Initialize a backend from a name, class, or instance"""
    logger.debug(f'Initializing backend: {backend} {cache_name}')

    # The 'cache_name' arg has a different purpose depending on the backend. If an equivalent
    # backend-specific keyword arg is specified, handle that here to avoid conflicts with the
    # 'cache_name' positional-or-keyword arg. In hindsight, a consistent positional-only or
    # keyword-only arg would have been better, but probably not worth a breaking change.
    cache_name_kwargs = [kwargs.pop(k) for k in CACHE_NAME_KWARGS if k in kwargs]
    cache_name = cache_name or cache_name_kwargs[0]

    # Already a backend instance
    if isinstance(backend, BaseCache):
        # Database names, file paths, etc. cannot be reliably updated after initialization, so
        # warn if the user passes both `cache_name` and a backend instance
        from ..policy import DEFAULT_CACHE_NAME

        if cache_name and cache_name != DEFAULT_CACHE_NAME:
            backend.cache_name = str(cache_name)
            warn(
                '`cache_name` cannot be set after backend initialization; '
                'please pass it to the backend class instead',
                DeprecationWarning,
                stacklevel=3,
            )
        return backend
    # If no backend is specified, use SQLite as default, unless the environment doesn't support it
    elif not backend:
        sqlite_supported = issubclass(BACKEND_CLASSES['sqlite'], BaseCache)
        backend = 'sqlite' if sqlite_supported else 'memory'

    # Get backend class by name
    backend = str(backend).lower()
    if backend not in BACKEND_CLASSES:
        raise ValueError(
            f'Invalid backend: {backend}. Provide a backend instance, or choose from one of the '
            f'following aliases: {list(BACKEND_CLASSES.keys())}'
        )
    return BACKEND_CLASSES[backend](cache_name, **kwargs)



"""Example of using requests-cache with the requests-mock library"""

from os.path import dirname, join
from unittest.mock import patch

import pytest
import requests
from requests import Session
from requests_mock import Adapter, NoMockAddress

from requests_cache import CachedSession

TEST_DB = join(dirname(__file__), 'httpbin_sample.test-db')
TEST_URLS = [
    'https://httpbin.org/get',
    'https://httpbin.org/html',
    'https://httpbin.org/json',
]
UNMOCKED_URL = 'https://httpbin.org/ip'


@pytest.fixture(scope='session')
def mock_session():
    """Fixture that provides a session with mocked URLs and responses based on cache data"""
    adapter = Adapter()
    cache = CachedSession(TEST_DB).cache

    for response in cache.responses.values():
        adapter.register_uri(
            response.request.method,
            response.request.url,
            content=response.content,
            headers=response.headers,
            status_code=response.status_code,
        )
        print(f'Added mock response: {response}')

    session = Session()
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    yield session


@patch.object(requests.adapters.HTTPAdapter, 'send', side_effect=ValueError('Real request made!'))
def test_mock_session(mock_http_adapter, mock_session):
    """Test that the mock_session fixture is working as expected"""
    # An error will be raised if a real request is made
    with pytest.raises(ValueError):
        requests.get(TEST_URLS[0])

    # All mocked URLs will return a response based on requests-cache data
    for url in TEST_URLS:
        response = mock_session.get(url)
        assert getattr(response, 'from_cache', False) is False

    # requests-mock will raise an error for an unmocked URL, as usual
    with pytest.raises(NoMockAddress):
        mock_session.get(UNMOCKED_URL)


def save_test_data():
    """Run once to save data to reuse for tests, for demo purposes.
    In practice, you could just run your application or tests with requests-cache installed.
    """
    session = CachedSession(TEST_DB)
    for url in TEST_URLS:
        session.get(url)


if __name__ == '__main__':
    save_test_data()



from datetime import datetime, timedelta
from io import BytesIO
from time import sleep

import pytest
from requests import Response
from urllib3.response import HTTPResponse

from requests_cache.models.response import CachedResponse, OriginalResponse, format_file_size
from requests_cache.policy import CacheActions, utcnow
from tests.conftest import MOCKED_URL


def test_basic_attrs(mock_session):
    response = CachedResponse.from_response(mock_session.get(MOCKED_URL))

    assert response.from_cache is True
    assert response.url == MOCKED_URL
    assert response.status_code == 200
    assert response.reason is None
    assert response.encoding == 'ISO-8859-1'
    assert response.headers['Content-Type'] == 'text/plain'
    assert response.text == 'mock response'
    assert isinstance(response.created_at, datetime)
    assert response.expires is None
    assert response.is_expired is False


@pytest.mark.parametrize(
    'response',
    [
        OriginalResponse(),
        OriginalResponse.wrap_response(Response(), CacheActions()),
    ],
)
def test_original_response(response):
    """Test both instantiating directly and wrapping a `requests.Response`"""
    response = OriginalResponse()
    assert response._content is False
    assert response._content_consumed is False
    assert response.status_code is None
    assert isinstance(response.created_at, datetime)
    assert response.expires is None
    assert response.revalidated is False
    assert response.cache_key == ''


def test_history(mock_session):
    original_response = mock_session.get(MOCKED_URL)
    original_response.history = [mock_session.get(MOCKED_URL)] * 3
    response = CachedResponse.from_response(original_response)
    assert len(response.history) == 3
    assert all(isinstance(r, CachedResponse) for r in response.history)


@pytest.mark.parametrize(
    'expires, is_expired',
    [
        (utcnow() + timedelta(days=1), False),
        (utcnow() - timedelta(days=1), True),
    ],
)
def test_is_expired(expires, is_expired, mock_session):
    response = CachedResponse.from_response(mock_session.get(MOCKED_URL), expires=expires)
    assert response.from_cache is True
    assert response.is_expired == is_expired


def test_is_expired__timezone_naive(mock_session):
    response = CachedResponse.from_response(
        mock_session.get(MOCKED_URL),
        expires=datetime.now() - timedelta(days=1),
    )
    assert response.is_expired is True


def test_iterator(mock_session):
    # Set up mock response with streamed content
    url = f'{MOCKED_URL}/stream'
    mock_raw_response = HTTPResponse(
        body=BytesIO(b'mock response'),
        status=200,
        request_method='GET',
        decode_content=False,
        preload_content=False,
    )
    mock_session.mock_adapter.register_uri(
        'GET',
        url,
        status_code=200,
        raw=mock_raw_response,
    )

    # Expect the same chunks of data from the original response and subsequent cached responses
    last_request_chunks = None
    for i in range(3):
        response = mock_session.get(url, stream=True)
        chunks = list(response.iter_lines())
        if i == 0:
            assert response.from_cache is False
        else:
            assert response.from_cache is True
            assert chunks == last_request_chunks
        last_request_chunks = chunks


def test_reset_expiration__extend_expiration(mock_session):
    # Start with an expired response
    response = CachedResponse.from_response(
        mock_session.get(MOCKED_URL),
        expires=utcnow() - timedelta(seconds=0.01),
    )
    assert response.is_expired is True

    # Set expiration in the future
    is_expired = response.reset_expiration(utcnow() + timedelta(seconds=0.01))
    assert is_expired is response.is_expired is False
    sleep(0.1)
    assert response.is_expired is True


def test_reset_expiration__shorten_expiration(mock_session):
    # Start with a non-expired response
    response = CachedResponse.from_response(
        mock_session.get(MOCKED_URL),
        expires=utcnow() + timedelta(seconds=1),
    )
    assert response.is_expired is False

    # Set expiration in the past
    is_expired = response.reset_expiration(utcnow() - timedelta(seconds=1))
    assert is_expired is response.is_expired is True


def test_size(mock_session):
    response = CachedResponse.from_response(mock_session.get(MOCKED_URL))
    response._content = None
    assert response.size == 0
    response._content = b'1' * 1024
    assert response.size == 1024


def test_str(mock_session):
    """Just ensure that a subset of relevant attrs get included in the response str; the format
    may change without breaking the test.
    """
    response = CachedResponse.from_response(mock_session.get(MOCKED_URL))
    response._content = b'1010'
    expected_values = ['GET', MOCKED_URL, 200, '4 bytes', 'created', 'expires', 'fresh']
    assert all(str(v) in str(response) for v in expected_values)


def test_repr(mock_session):
    """Just ensure that a subset of relevant attrs get included in the response repr"""
    response = CachedResponse.from_response(mock_session.get(MOCKED_URL))
    expected_values = ['GET', MOCKED_URL, 200, 'ISO-8859-1', response.headers]
    print(repr(response))
    assert repr(response).startswith('CachedResponse(') and repr(response).endswith(')')
    assert all(str(v) in repr(response) for v in expected_values)


@pytest.mark.parametrize(
    'n_bytes, expected_size',
    [
        (None, '0 bytes'),
        (5, '5 bytes'),
        (3 * 1024, '3.00 KiB'),
        (1024 * 3000, '2.93 MiB'),
        (1024 * 1024 * 5000, '4.88 GiB'),
    ],
)
def test_format_file_size(n_bytes, expected_size):
    assert format_file_size(n_bytes) == expected_size



#!/usr/bin/env python
"""
An example of loading CachedSession settings from an external config file.

Limitations:
* Does not include backend or serializer settings
* Does not include settings specified as python expressions, for example `timedelta` objects or
  callback functions
"""
from pathlib import Path

import yaml

from requests_cache import CachedSession, CacheSettings

CONFIG_FILE = Path(__file__).parent / 'external_config.yml'


def load_settings() -> CacheSettings:
    """Load settings from a YAML config file"""
    with open(CONFIG_FILE) as f:
        settings = yaml.safe_load(f)
    return CacheSettings(**settings['cache_settings'])


if __name__ == '__main__':
    session = CachedSession()
    session.settings = load_settings()
    print('Loaded settings:\n', session.settings)



from collections import OrderedDict
from decimal import Decimal
from unittest.mock import patch

import pytest

from requests_cache.backends import DynamoDbCache, DynamoDbDict
from tests.conftest import CACHE_NAME, fail_if_no_connection
from tests.integration.base_cache_test import BaseCacheTest
from tests.integration.base_storage_test import BaseStorageTest

AWS_OPTIONS = {
    'endpoint_url': 'http://localhost:8000',
    'region_name': 'us-east-1',
    'aws_access_key_id': 'placeholder',
    'aws_secret_access_key': 'placeholder',
}


@pytest.fixture(scope='module', autouse=True)
@fail_if_no_connection(connect_timeout=5)
def ensure_connection():
    """Fail all tests in this module if DynamoDB is not running"""
    import boto3

    client = boto3.client('dynamodb', **AWS_OPTIONS)
    client.describe_limits()


class TestDynamoDbDict(BaseStorageTest):
    storage_class = DynamoDbDict
    init_kwargs = AWS_OPTIONS

    def init_cache(self, cache_name=CACHE_NAME, index=0, clear=True, **kwargs):
        """For tests that use multiple tables, make index part of the table name"""
        kwargs = {**self.init_kwargs, **kwargs}
        cache = self.storage_class(f'{cache_name}_{index}', **kwargs)
        if clear:
            cache.clear()
        return cache

    @patch('requests_cache.backends.dynamodb.boto3.resource')
    def test_connection_kwargs(self, mock_resource):
        """A spot check to make sure optional connection kwargs gets passed to connection"""
        DynamoDbDict('test_table', region_name='us-east-2', invalid_kwarg='???')
        mock_resource.assert_called_with('dynamodb', region_name='us-east-2')

    @patch('requests_cache.backends.dynamodb.boto3.resource')
    def test_no_create_table(self, mock_resource):
        DynamoDbDict('test_table', region_name='us-east-2', create_table=False)
        assert mock_resource.create_table.call_count == 0

    def test_enable_ttl_error(self):
        """An error other than 'table already exists' should be reraised"""
        from botocore.exceptions import ClientError

        cache = self.init_cache()
        error = ClientError({'Error': {'Code': 'NullPointerException'}}, 'CreateTable')
        with patch.object(cache.connection.meta.client, 'update_time_to_live', side_effect=error):
            with pytest.raises(ClientError):
                cache._enable_ttl()

    def test_create_table_error(self):
        """An error other than 'ttl already enabled' should be reraised"""
        from botocore.exceptions import ClientError

        cache = self.init_cache()
        error = ClientError({'Error': {'Code': 'NullPointerException'}}, 'CreateTable')
        with patch.object(cache.connection, 'create_table', side_effect=error):
            with pytest.raises(ClientError):
                cache._create_table()

    @pytest.mark.parametrize('ttl_enabled', [True, False])
    def test_ttl(self, ttl_enabled):
        """DynamoDB's TTL removal process can take up to 48 hours to run, so just test if the
        'ttl' attribute is set correctly if enabled, and not set if disabled.
        """
        cache = self.init_cache(ttl=ttl_enabled)
        item = OrderedDict(foo='bar')
        item.expires_unix = 60
        cache['key'] = item

        # 'ttl' is a reserved word, so to retrieve it we need to alias it
        item = cache._table.get_item(
            Key={'key': 'key'},
            ProjectionExpression='#t',
            ExpressionAttributeNames={'#t': 'ttl'},
        )
        ttl_value = item['Item'].get('ttl')

        if ttl_enabled:
            assert isinstance(ttl_value, Decimal)
        else:
            assert ttl_value is None


class TestDynamoDbCache(BaseCacheTest):
    backend_class = DynamoDbCache
    init_kwargs = AWS_OPTIONS



from sys import version_info

import pytest

from requests_cache.backends import BaseCache, DictStorage
from tests.integration.base_cache_test import BaseCacheTest
from tests.integration.base_storage_test import BaseStorageTest


class TestMemoryDict(BaseStorageTest):
    storage_class = DictStorage

    def init_cache(self, clear=True, **kwargs):
        cache = self.storage_class(**kwargs)
        if clear:
            cache.clear()
        return cache

    def test_same_settings(self):
        """This test from base class doesn't apply here"""


class TestMemoryCache(BaseCacheTest):
    backend_class = BaseCache

    # TODO: Remove when fixed
    def test_response_no_duplicate_read(self):
        if version_info >= (3, 12):
            pytest.xfail('Known (very minor) bug with python 3.12')
        super().test_response_no_duplicate_read()



import os
from shutil import rmtree
from tempfile import gettempdir
from threading import RLock
from time import sleep

import pytest
from platformdirs import user_cache_dir

from requests_cache.backends import FileCache, FileDict, LRUFileDict
from requests_cache.backends.filesystem import LRUDict
from requests_cache.serializers import (
    SERIALIZERS,
    SerializerPipeline,
    Stage,
    json_serializer,
    safe_pickle_serializer,
    utf8_serializer,
    yaml_serializer,
)
from tests.conftest import CACHE_NAME, HTTPBIN_FORMATS, HTTPBIN_METHODS
from tests.integration.base_cache_test import BaseCacheTest
from tests.integration.base_storage_test import BaseStorageTest

# Handle optional dependencies if they're not installed,
# so any skipped tests will explicitly be shown in pytest output
TEST_SERIALIZERS = SERIALIZERS.copy()
try:
    TEST_SERIALIZERS['safe_pickle'] = safe_pickle_serializer(secret_key='hunter2')
except ImportError:
    TEST_SERIALIZERS['safe_pickle'] = 'safe_pickle_placeholder'


def _valid_serializer(serializer) -> bool:
    return isinstance(serializer, (SerializerPipeline, Stage))


class TestFileDict(BaseStorageTest):
    storage_class = FileDict
    picklable = True

    @classmethod
    def teardown_class(cls):
        rmtree(CACHE_NAME, ignore_errors=True)

    def init_cache(self, index=0, clear=True, **kwargs) -> FileDict:
        cache = self.storage_class(f'{CACHE_NAME}_{index}', use_temp=True, **kwargs)
        if clear:
            cache.clear()
        return cache

    def test_use_cache_dir(self):
        relative_path = self.storage_class(CACHE_NAME).cache_dir
        cache_dir_path = self.storage_class(CACHE_NAME, use_cache_dir=True).cache_dir
        assert not str(relative_path).startswith(user_cache_dir())
        assert str(cache_dir_path).startswith(user_cache_dir())

    def test_use_temp(self):
        relative_path = self.storage_class(CACHE_NAME).cache_dir
        temp_path = self.storage_class(CACHE_NAME, use_temp=True).cache_dir
        assert not str(relative_path).startswith(gettempdir())
        assert str(temp_path).startswith(gettempdir())

    def test_custom_extension(self):
        cache = self.init_cache(extension='dat')
        cache['key'] = 'value'
        assert cache._key2path('key').suffix == '.dat'

    def test_size(self):
        """Check that size updates with bytes added/updated/removed."""
        cache = self.init_cache(plaintext=True)
        assert cache.size() == 0

        # Create a 1KB file
        file_content = '0' * 1024
        cache['key'] = file_content
        assert cache.size() == 1026  # +2 for UTF-8 BOM

        cache.clear()
        assert cache.size() == 0


class TestLRUFileDict(TestFileDict):
    """Test the LRUFileDict with the same tests as the FileDict."""

    storage_class = LRUFileDict

    def init_cache(
        self,
        index=0,
        clear=True,
        max_cache_bytes=10000,
        block_bytes=1,
        plaintext=False,
        **kwargs,
    ) -> LRUFileDict:
        """Initialize a LRUFileDict with common test parameters.

        Args:
            shared_cache_name: If provided, use this exact cache name (for multi-process simulation).
                              If None, create a unique cache name for isolation.
        """
        if plaintext:
            kwargs['serializer'] = None

        cache = self.storage_class(
            f'{CACHE_NAME}_{index}',
            use_temp=True,
            max_cache_bytes=max_cache_bytes,
            block_bytes=block_bytes,
            **kwargs,
        )
        if clear:
            cache.clear()
        return cache

    @pytest.mark.parametrize('block_bytes', [0, -1, 101])
    def test_init__invalid_block_bytes(self, block_bytes):
        """This block size is invalid."""
        with pytest.raises(ValueError):
            self.init_cache(
                max_file_bytes=100,
                max_cache_bytes=100,
                block_bytes=block_bytes,
            )

    def test_init__invalid_max_file_size(self):
        """Cannot have file size greater than maximum cache size."""
        with pytest.raises(ValueError):
            self.init_cache(
                max_file_bytes=1000,
                max_cache_bytes=100,
                block_bytes=100,
            )

    def test_init__max_file_bytes_defaults_to_max_size(self):
        """The argument is initialized properly."""
        lfd = self.init_cache(max_cache_bytes=100, block_bytes=100)
        assert lfd.max_file_bytes == lfd.max_cache_bytes == 100

    def test_init__shared_lock(self):
        """Test passing the lock parameter."""
        lock = RLock()
        cache1 = self.init_cache(lock=lock)
        cache2 = self.init_cache(lock=lock)
        cache3 = self.init_cache()
        assert cache1.lock is cache2.lock
        assert cache1.lock is not cache3.lock

    def test_multiple_cache_instances(self):
        """Multiple cache instances should see the same files."""
        lfd = self.init_cache()
        lfd2 = self.init_cache()

        # Adding items
        lfd['key'] = 'value'
        assert 'key' in lfd
        assert 'key' in lfd2

        # Replacing items
        lfd['key'] = 'new_value'
        assert lfd2['key'] == 'new_value'

        # Removing items
        del lfd['key']
        assert 'key' not in lfd2

    @pytest.mark.parametrize('max_cache_bytes', [1000, 10000])
    def test_do_not_store_files_too_big(self, max_cache_bytes: int):
        """If a file is too big, it should not be cached at all."""
        lfd = self.init_cache(max_cache_bytes=max_cache_bytes)
        file_content = '0' * (max_cache_bytes + 1)
        lfd['key'] = file_content
        assert 'key' not in lfd.keys()

    @pytest.mark.parametrize('files_on_disk', [3, 100])
    def test_evict_oldest_files(self, files_on_disk: int):
        """Check that the files are added up to the size.

        After adding, the least recently used files are removed.
        """
        max_cache_bytes = 1000
        lfd = self.init_cache(max_cache_bytes=max_cache_bytes, plaintext=True)
        file_content = '0' * (max_cache_bytes // files_on_disk)
        for i in range(files_on_disk):
            lfd[f'key_{i}'] = file_content
        assert len(list(lfd.paths())) == files_on_disk, (
            f'Expecting {files_on_disk} files with size '
            f'{len(file_content)}, got {len(list(lfd.paths()))}: '
            f'{sorted(lfd.paths())}'
        )
        assert len(lfd) == files_on_disk
        lfd['new_file'] = file_content
        assert len(list(lfd.paths())) == files_on_disk, 'One file should be dropped.'
        assert 'key_0' not in lfd.keys(), 'File key_0 should be dropped because it is the last one.'

    def test_evict__already_deleted(self):
        """Test behahior when a file is being evicted but has already been (manually) deleted on disk"""
        lfd = self.init_cache(max_cache_bytes=1000, plaintext=True)
        lfd['key_0'] = '0' * 500
        sleep(0.01)
        lfd['key_1'] = '0' * 500

        # Manually delete file, so LRU index is now out of sync
        os.unlink(lfd._key2path('key_0'))
        assert 'key_0' not in lfd.keys()

        # Adding this key should trigger eviction
        lfd['key_2'] = '0' * 500
        assert len(lfd) == 2
        assert len(lfd.lru_index) == 2
        assert 'key_0' not in lfd.lru_index

    def test_sync_index(self):
        """With sync_index=True, the LRU index should be updated on init for any manual file changes."""
        lfd = self.init_cache(max_cache_bytes=1000, plaintext=True)
        lfd['key_0'] = '0' * 100
        sleep(0.01)
        lfd['key_1'] = '0' * 100

        # Manually delete a file outside of cache interface
        os.unlink(lfd._key2path('key_0'))

        # Index should be updated on init
        lfd = self.init_cache(max_cache_bytes=1000, sync_index=True, plaintext=True, clear=False)
        lfd._evict(100)
        assert len(lfd) == 1
        assert len(lfd.lru_index) == 1

    def test_large_file_evicts_all(self):
        """If we add a big file, we should delete everything."""
        lfd = self.init_cache(max_cache_bytes=1000, plaintext=True)
        for i in range(100):
            lfd[f'key_{i}'] = '1' * 10
        lfd['big_file'] = '0' * 1000
        assert lfd.keys() == ['big_file']

    def test_size(self):
        """Check that size updates with bytes added/updated/removed."""
        cache = self.init_cache(plaintext=True)
        assert cache.size() == 0
        assert len(cache.keys()) == 0
        assert len(cache) == 0

        # Adding items
        cache['key'] = 'value'
        assert cache.size() == len('value')
        cache['key2'] = 'value2'
        assert cache.size() == len('value') + len('value2')

        # Replacing items
        cache['key'] = 'new_value'
        assert cache.size() == len('new_value') + len('value2')

        # Removing items
        del cache['key2']
        assert cache.size() == len('new_value')

    @pytest.mark.parametrize(
        ('block_bytes', 'file_size', 'expected_size'),
        [
            (1, 10, 10),
            (10, 9, 10),
            (10, 10, 10),
            (10, 11, 20),
            (10, 12, 20),
            (1024, 3, 1024),
            (100, 1111, 1200),
        ],
    )
    def test_get_size_on_disk(self, block_bytes, file_size, expected_size):
        """Compute the block size."""
        cache = self.init_cache(block_bytes=block_bytes)
        assert cache._get_size_on_disk(file_size) == expected_size
        assert cache._get_size_on_disk(-file_size) == -expected_size

    @pytest.mark.parametrize('block_bytes', [1, 10, 128])
    def test_block_alignment(self, block_bytes: int):
        """File size should be aligned with the block on the file system, and be taken into account
        when adding, replacing, and deleting items.
        """
        block_lfd = self.init_cache(block_bytes=block_bytes, plaintext=True)

        # Adding items
        block_lfd['file'] = '0' * 100
        assert block_lfd.size() % block_lfd.block_bytes == 0
        assert block_lfd.size() == block_lfd._get_size_on_disk(100)

        # Replacing items
        block_lfd['file'] = '0' * 123
        assert block_lfd.size() % block_lfd.block_bytes == 0
        assert block_lfd.size() == block_lfd._get_size_on_disk(123)

        # Deleting items
        del block_lfd['file']
        block_lfd['key1'] = '2' * 100
        size_1 = block_lfd.size()
        block_lfd['key2'] = '2' * 123
        size_1_2 = block_lfd.size()
        assert size_1_2 > size_1
        assert size_1_2 % block_lfd.block_bytes == 0
        del block_lfd['key1']
        size_2 = size_1_2 - size_1
        assert block_lfd.size() == size_2
        assert size_2 % block_lfd.block_bytes == 0
        del block_lfd['key2']
        assert block_lfd.size() == 0

    @pytest.mark.parametrize('max_cache_bytes', [5, 6, 7, 8, 9])
    def test_clear_with_unaligned_blocks(self, max_cache_bytes):
        lfd = self.init_cache(
            max_file_bytes=5,
            max_cache_bytes=max_cache_bytes,
            block_bytes=5,
        )
        lfd.clear()
        lfd['x'] = '123'
        assert lfd.size() == 5
        lfd['y'] = ''
        assert lfd.size() == 5
        assert 'x' not in lfd, 'x must have been deleted'


class TestLRUDict:
    def init_cache(self, **kwargs):
        cache = LRUDict(CACHE_NAME, table_name='lru', use_temp=True, **kwargs)
        cache.clear()
        return cache

    @classmethod
    def teardown_class(cls):
        try:
            os.unlink(f'{CACHE_NAME}.sqlite')
        except Exception:
            pass

    def test_get_set(self):
        cache = self.init_cache()
        cache['key1'] = 100
        cache['key2'] = 200
        cache['key3'] = 300
        assert cache['key1'] == 100
        assert cache['key2'] == 200
        assert cache['key3'] == 300

        with pytest.raises(KeyError):
            _ = cache['nonexistent']

    def test_delete(self):
        cache = self.init_cache()
        cache['key'] = 0

        del cache['key']
        with pytest.raises(KeyError):
            _ = cache['key']

    def test_count(self):
        cache = self.init_cache()
        assert cache.count() == 0

        cache['key1'] = 100
        assert cache.count() == 1

        cache['key2'] = 200
        cache['key3'] = 300
        assert cache.count() == 3

        del cache['key1']
        assert cache.count() == 2

    def test_clear(self):
        cache = self.init_cache()
        cache['key1'] = 100
        cache['key2'] = 200
        cache['key3'] = 300
        assert len(cache) == 3
        assert cache.total_size() == 600

        cache.clear()
        assert len(cache) == 0
        assert cache.total_size() == 0

    def test_get_lru(self):
        cache = self.init_cache()
        assert cache.get_lru(total_size=1) == []

        cache['key0'] = 0
        cache['key1'] = 100
        cache['key2'] = 200
        cache['key3'] = 300
        sleep(0.001)
        cache.update_access_time('key1')

        # Order should be (from least to most recent): key0, key2, key3, key1
        assert cache.get_lru(total_size=1) == ['key0', 'key2']
        assert cache.get_lru(total_size=200) == ['key0', 'key2']
        assert cache.get_lru(total_size=201) == ['key0', 'key2', 'key3']
        assert cache.get_lru(total_size=500) == ['key0', 'key2', 'key3']
        assert cache.get_lru(total_size=501) == ['key0', 'key2', 'key3', 'key1']
        assert cache.get_lru(total_size=600) == ['key0', 'key2', 'key3', 'key1']
        assert cache.get_lru(total_size=700) == ['key0', 'key2', 'key3', 'key1']

    @pytest.mark.parametrize(
        ('sort_key', 'reversed_order', 'limit', 'expected'),
        [
            # Sort by access_time (default)
            ('access_time', False, None, ['key1', 'key2', 'key3']),
            ('access_time', True, None, ['key3', 'key2', 'key1']),
            ('access_time', False, 2, ['key1', 'key2']),
            ('access_time', True, 2, ['key3', 'key2']),
            # Sort by size
            ('size', False, None, ['key1', 'key2', 'key3']),
            ('size', True, None, ['key3', 'key2', 'key1']),
            ('size', False, 2, ['key1', 'key2']),
            # Sort by key (alphabetical)
            ('key', False, None, ['key1', 'key2', 'key3']),
            ('key', True, None, ['key3', 'key2', 'key1']),
        ],
    )
    def test_sorted(self, sort_key, reversed_order, limit, expected):
        cache = self.init_cache()

        # Add items with different sizes and access times
        cache['key1'] = 100
        sleep(0.001)
        cache['key2'] = 200
        sleep(0.001)
        cache['key3'] = 300

        result = list(cache.sorted(key=sort_key, reversed=reversed_order, limit=limit))
        assert result == expected

    def test_sorted__invalid_key(self):
        cache = self.init_cache()
        cache['key1'] = 100

        with pytest.raises(ValueError, match='Invalid sort key: invalid'):
            list(cache.sorted(key='invalid'))

    def test_total_size(self):
        """Test that total cache size is tracked correctly"""
        cache = self.init_cache()
        assert cache.total_size() == 0

        # Create
        cache['key1'] = 100
        assert cache.total_size() == 100
        cache['key2'] = 200
        assert cache.total_size() == 300

        # Update
        cache['key1'] = 150
        assert cache.total_size() == 350

        # Delete
        del cache['key1']
        assert cache.total_size() == 200

    def test_update_access_time(self):
        cache = self.init_cache()
        cache['key1'] = 100
        cache['key2'] = 200
        cache['key3'] = 300
        assert list(cache.sorted()) == ['key1', 'key2', 'key3']

        sleep(0.001)
        cache.update_access_time('key1')
        assert list(cache.sorted()) == ['key2', 'key3', 'key1']

    def test_update_access_time__keyerror(self):
        cache = self.init_cache()
        with pytest.raises(KeyError):
            cache.update_access_time('nonexistent')


class TestFileCache(BaseCacheTest):
    backend_class = FileCache
    init_kwargs = {'use_temp': True}

    @pytest.mark.parametrize('serializer', TEST_SERIALIZERS.values())
    @pytest.mark.parametrize('method', HTTPBIN_METHODS)
    @pytest.mark.parametrize('field', ['params', 'data', 'json'])
    def test_all_methods(self, field, method, serializer):
        """Test all relevant combinations of methods X data fields X serializers"""
        if not _valid_serializer(serializer):
            pytest.skip(f'Dependencies not installed for {serializer}')
        super().test_all_methods(field, method, serializer)

    @pytest.mark.parametrize('serializer', TEST_SERIALIZERS.values())
    @pytest.mark.parametrize('response_format', HTTPBIN_FORMATS)
    def test_all_response_formats(self, response_format, serializer):
        """Test all relevant combinations of response formats X serializers"""
        if not _valid_serializer(serializer):
            pytest.skip(f'Dependencies not installed for {serializer}')
        serializer.set_decode_content(False)
        super().test_all_response_formats(response_format, serializer)

    @pytest.mark.parametrize('serializer', [json_serializer, yaml_serializer])
    @pytest.mark.parametrize('response_format', HTTPBIN_FORMATS)
    def test_all_response_formats__no_decode_content(self, response_format, serializer):
        """Test with decode_content=True for text-based serialization formats"""
        if not _valid_serializer(serializer):
            pytest.skip(f'Dependencies not installed for {serializer}')
        serializer.set_decode_content(True)
        self.test_all_response_formats(response_format, serializer)

    @pytest.mark.parametrize('serializer_name', SERIALIZERS.keys())
    def test_paths(self, serializer_name):
        if not isinstance(SERIALIZERS[serializer_name], SerializerPipeline):
            pytest.skip(f'Dependencies not installed for {serializer_name}')

        session = self.init_session(serializer=serializer_name)
        num_files = 20
        for i in range(num_files):
            session.cache.responses[f'key_{i}'] = {f'value_{i}': i}

        expected_extension = serializer_name.replace('pickle', 'pkl')
        if 'json' in serializer_name:
            expected_extension = 'json'
        assert len(list(session.cache.paths())) == num_files
        for path in session.cache.paths():
            assert path.is_file()
            assert path.suffix == f'.{expected_extension}'

        # Redirects db should be in the same directory as response files
        assert session.cache.redirects.db_path.parent == session.cache.responses.cache_dir

    @pytest.mark.parametrize('max_cache_bytes', [1000, 10000])
    @pytest.mark.parametrize('files_on_disk', [3, 100])
    def test_drop_oldest_files(self, max_cache_bytes, files_on_disk):
        """Check that the files are added up to the size.

        After adding, the first ones are dropped.
        """
        file_content = '0' * (max_cache_bytes // files_on_disk - 2)  # remove UTF8 BOM
        session = self.init_session(max_cache_bytes=max_cache_bytes, block_bytes=1)
        for i in range(files_on_disk):
            session.cache.responses[f'key_{i}'] = file_content
        assert len(list(session.cache.paths())) == files_on_disk, (
            f'Expecting {files_on_disk} files with size {len(file_content)}, got {len(list(session.cache.paths()))}'
        )
        assert len(list(session.cache.paths())) == files_on_disk
        session.cache.responses['new_file'] = file_content
        assert len(list(session.cache.paths())) == files_on_disk, 'One file should be dropped.'
        assert 'key_0' not in session.cache.responses.keys(), (
            'File key_0 should be dropped because it is the last one.'
        )

    @pytest.mark.parametrize('max_size_on_disk', [1000, 10000])
    def test_do_not_store_files_too_big(self, max_size_on_disk):
        """If a file is too big, it should not be cached at all."""
        file_content = '0' * (max_size_on_disk + 1)
        session = self.init_session(max_cache_bytes=max_size_on_disk, block_bytes=1)
        session.cache.responses['key'] = file_content
        assert 'key' not in session.cache.responses.keys()

    def test_size(self):
        """Check that the total bytes are summed up if files are added."""
        session = self.init_session(max_cache_bytes=100, block_bytes=1, serializer=utf8_serializer)
        start_bytes = session.cache.responses.size()
        session.cache.responses['key_1'] = 'value_1'
        assert session.cache.responses.size() == start_bytes + len('value_1')



from unittest.mock import patch

import pytest
from redis import StrictRedis

from requests_cache.backends import RedisCache, RedisDict, RedisHashDict
from tests.conftest import fail_if_no_connection, httpbin
from tests.integration.base_cache_test import BaseCacheTest
from tests.integration.base_storage_test import BaseStorageTest


@pytest.fixture(scope='module', autouse=True)
@fail_if_no_connection()
def ensure_connection():
    """Fail all tests in this module if Redis is not running"""
    from redis import Redis

    client = Redis()
    try:
        client.info()
    finally:
        client.close()


class TestRedisDict(BaseStorageTest):
    storage_class = RedisDict
    num_instances = 1  # Only supports a single instance, since it stores items under top-level keys

    @patch('requests_cache.backends.redis.StrictRedis')
    def test_connection_kwargs(self, mock_redis):
        """A spot check to make sure optional connection kwargs get passed to connection"""
        RedisCache('test', username='user', password='pass', invalid_kwarg='???')
        mock_redis.assert_called_with(username='user', password='pass')


class TestRedisHashDict(TestRedisDict):
    storage_class = RedisHashDict
    num_instances: int = 10  # Supports multiple instances, since this stores items under hash keys
    picklable = True
    init_kwargs = {'serializer': 'pickle'}


class TestRedisCache(BaseCacheTest):
    backend_class = RedisCache

    @patch.object(StrictRedis, 'setex')
    def test_ttl(self, mock_setex):
        session = self.init_session(expire_after=60)
        session.get(httpbin('get'))
        call_args = mock_setex.mock_calls[0][1]
        assert call_args[1] == 3660  # Should be expiration + default offset

    @patch.object(StrictRedis, 'setex')
    def test_ttl__offset(self, mock_setex):
        session = self.init_session(expire_after=60, ttl_offset=500)
        session.get(httpbin('get'))
        call_args = mock_setex.mock_calls[0][1]
        assert call_args[1] == 560  # Should be expiration + custom offset

    @patch.object(StrictRedis, 'setex')
    @patch.object(StrictRedis, 'set')
    def test_ttl__disabled(self, mock_set, mock_setex):
        session = self.init_session(expire_after=60, ttl=False)
        session.get(httpbin('get'))
        mock_setex.assert_not_called()
        mock_set.assert_called()



"""Data models used to serialize response data"""

# ruff: noqa: F401
from typing import Union

from requests import PreparedRequest, Request, Response

from .base import RichMixin
from .raw_response import CachedHTTPResponse
from .request import CachedRequest
from .response import CachedResponse, DecodedContent, OriginalResponse

AnyResponse = Union[OriginalResponse, CachedResponse]
AnyRequest = Union[Request, PreparedRequest, CachedRequest]
AnyPreparedRequest = Union[PreparedRequest, CachedRequest]



#!/usr/bin/env python
"""
Example utilities to export responses to a format compatible with VCR-based libraries, including:
* [vcrpy](https://github.com/kevin1024/vcrpy)
* [betamax](https://github.com/betamaxpy/betamax)
"""
from importlib.metadata import version as pkg_version
from os import makedirs
from os.path import abspath, dirname, expanduser, join
from typing import Any, Dict, Iterable
from urllib.parse import urlparse

import yaml

from requests_cache import BaseCache, CachedResponse, CachedSession
from requests_cache.serializers.preconf import yaml_preconf_stage


def to_vcr_cassette(cache: BaseCache, path: str):
    """Export cached responses to a VCR-compatible YAML file (cassette)

    Args:
        cache: Cache instance containing response data to export
        path: Path for new cassette file
    """

    responses = cache.responses.values()
    write_cassette(to_vcr_cassette_dict(responses), path)


def to_vcr_cassettes_by_host(cache: BaseCache, cassette_dir: str = '.'):
    """Export cached responses as VCR-compatible YAML files (cassettes), split into separate files
    based on request host

    Args:
        cache: Cache instance containing response data to export
        cassette_dir: Base directory for cassette library
    """
    responses = cache.responses.values()
    for host, cassette in to_vcr_cassette_dicts_by_host(responses).items():
        write_cassette(cassette, join(cassette_dir, f'{host}.yml'))


def to_vcr_cassette_dict(responses: Iterable[CachedResponse]) -> Dict:
    """Convert responses to a VCR cassette dict"""
    return {
        'http_interactions': [to_vcr_episode(r) for r in responses],
        'recorded_with': f'requests-cache {pkg_version("requests_cache")}',
    }


def to_vcr_episode(response: CachedResponse) -> Dict:
    """Convert a single response to a VCR-compatible response ("episode") dict"""
    # Do most of the work with cattrs + default YAML conversions
    response_dict = yaml_preconf_stage.dumps(response)

    def _to_multidict(d):
        return {k: [v] for k, v in d.items()}

    # Translate requests.Response structure into VCR format
    return {
        'request': {
            'body': response_dict['request']['body'],
            'headers': _to_multidict(response_dict['request']['headers']),
            'method': response_dict['request']['method'],
            'uri': response_dict['request']['url'],
        },
        'response': {
            'body': {'string': response_dict['_content'], 'encoding': response_dict['encoding']},
            'headers': _to_multidict(response_dict['headers']),
            'status': {'code': response_dict['status_code'], 'message': response_dict['reason']},
            'url': response_dict['url'],
        },
        'recorded_at': response_dict['created_at'],
    }


def to_vcr_cassette_dicts_by_host(responses: Iterable[CachedResponse]) -> Dict[str, Dict]:
    responses_by_host: Dict[str, Any] = {}
    for response in responses:
        host = urlparse(response.request.url).netloc
        responses_by_host.setdefault(host, [])
        responses_by_host[host].append(response)
    return {host: to_vcr_cassette_dict(responses) for host, responses in responses_by_host.items()}


def write_cassette(cassette, path):
    path = abspath(expanduser(path))
    makedirs(dirname(path), exist_ok=True)
    with open(path, 'w') as f:
        f.write(yaml.safe_dump(cassette))


# Create an example cache and export it to a cassette
if __name__ == '__main__':
    cache_dir = 'example_cache'
    session = CachedSession(join(cache_dir, 'http_cache.sqlite'))
    session.get('https://httpbin.org/get')
    session.get('https://httpbin.org/json')
    to_vcr_cassette(session.cache, join(cache_dir, 'http_cache.yaml'))



#!/usr/bin/env python
# flake8: noqa: F841
"""
A simple example using requests-cache with [httpbin](https://httpbin.org)
"""
import time

from requests_cache import CachedSession


def main():
    session = CachedSession('example_cache', backend='sqlite')

    # The real request will only be made once; afterward, the cached response is used
    for _ in range(5):
        response = session.get('https://httpbin.org/get')

    # This is more obvious when calling a slow endpoint
    for _ in range(5):
        response = session.get('https://httpbin.org/delay/2')

    # Caching can be disabled if we want to get a fresh page and not cache it
    with session.cache_disabled():
        print(session.get('https://httpbin.org/ip').text)

    # Get some debugging info about the cache
    print(session.cache)
    print('Cached URLS:')
    print('\n'.join(session.cache.urls()))


if __name__ == '__main__':
    t = time.time()
    main()
    print('Elapsed: %.3f seconds' % (time.time() - t))



#!/usr/bin/env python
"""
This example displays a graph of request rates over time. Requests are continuously sent to URLs
randomly picked from a fixed number of possible URLs. This demonstrates how average request rate
increases as the proportion of cached requests increases.

Try running this example with different cache settings and URLs to see how the graph changes.
"""
from random import randint
from time import time

from rich.live import Live
from rich.progress import BarColumn, MofNCompleteColumn, Progress
from rich.table import Table

from requests_cache import CachedSession

N_UNIQUE_REQUESTS = 200


class RPSProgress(Progress):
    """Display a bar chart of requests per second"""

    def __init__(self, interval: int = 1, scale: int = 500, **kwargs):
        super().__init__(BarColumn(), '{task.completed}', **kwargs)
        self.current_task = None
        self.interval = interval
        self.interval_start = None
        self.scale = scale
        self.total_requests = 0
        self.next_interval()

    def next_interval(self):
        """Create a new task to draw the next line on the bar chart"""
        self.current_task = self.add_task('barchart_line', total=self.scale)
        self.interval_start = time()

    def count_request(self):
        if time() - self.interval_start >= self.interval:
            self.next_interval()
        self.advance(self.current_task)
        self.total_requests += 1


class CacheRPSProgress:
    """Track requests per second plus cache size in a single live view"""

    def __init__(self, n_unique_requests: int = 100):
        self.rps_progress = RPSProgress()
        self.cache_progress = Progress(
            BarColumn(complete_style='blue'),
            '[cyan]Requests cached:',
            MofNCompleteColumn(),
        )
        header = Progress(BarColumn(), '[cyan]Requests per second')
        header.add_task('')
        self.cache_task = self.cache_progress.add_task('', total=n_unique_requests)
        self.n_unique_requests = n_unique_requests
        self.start_time = time()

        self.table = Table.grid()
        self.table.add_row(header)
        self.table.add_row(self.rps_progress)
        self.table.add_row(self.cache_progress)
        self.live = Live(self.table, refresh_per_second=10)

    def __enter__(self):
        """Start live view on ctx enter"""
        self.live.__enter__()
        self.log(
            '[cyan]Measuring request rate with '
            f'[white]{self.n_unique_requests}[cyan] total unique requests'
        )
        self.log('[cyan]Press [white]Ctrl+C[cyan] to exit')
        return self

    def __exit__(self, *args):
        """Show stats on ctx exit"""
        self.live.__exit__(*args)
        elapsed = time() - self.start_time
        self.log(
            f'[cyan]Sent a total of [white]{self.total_requests}[cyan] '
            f'requests in [white]{elapsed:.2f}[cyan] seconds '
        )

        self.log(f'[cyan]Average: [white]{int(self.total_requests/elapsed)}[cyan] requests/second')

    @property
    def total_requests(self):
        return self.rps_progress.total_requests

    def count_request(self):
        self.rps_progress.count_request()

    def update_cache_size(self, size: int):
        self.cache_progress.update(self.cache_task, completed=size)

    def log(self, msg: str):
        self.cache_progress.log(msg)


def test_rps(session):
    session.cache.clear()

    # Send a request to one of a fixed number of unique URLs
    def random_request():
        request_number = randint(1, N_UNIQUE_REQUESTS)
        session.get(f'https://httpbin.org/get?page={request_number}')

    # Show request rate over time and total cached (unexpired) requests
    with CacheRPSProgress(N_UNIQUE_REQUESTS) as progress:
        while True:
            try:
                random_request()
                progress.count_request()
                progress.update_cache_size(session.cache.responses.count(expired=False))
            except KeyboardInterrupt:
                break


if __name__ == '__main__':
    session = CachedSession(use_temp=True, expire_after=30)
    test_rps(session)



"""An optional wrapper for common commands used in local development and CI.

TODO: This will likely be replaced by uv tasks, when released:
https://github.com/astral-sh/uv/issues/5903

Notes:
* 'test' and 'test-<python version>' commands: nox will create separate virtualenvs per python
  version
* 'lint' command: tools and environments are managed by pre-commit
* All other commands: the current environment will be used instead of creating new ones
    * If using uv to manage virtualenvs, manually activate it first with `source .venv/bin/activate`
* Run `nox -l` to see all available commands
* See Contributing Guide for more usage details:
  https://github.com/requests-cache/requests-cache/blob/main/CONTRIBUTING.md
"""

import platform
from os import getenv
from os.path import join
from pathlib import Path
from shutil import rmtree

import nox

nox.options.reuse_existing_virtualenvs = True
nox.options.sessions = ['lint', 'cov']

LIVE_DOCS_PORT = 8181
LIVE_DOCS_IGNORE = ['*.pyc', '*.tmp', join('**', 'modules', '*')]
LIVE_DOCS_WATCH = ['requests_cache', 'examples']

DOCS_DIR = Path('docs')
DOC_BUILD_DIR = DOCS_DIR / '_build' / 'html'
TEST_DIR = Path('tests')
CLEAN_DIRS = ['dist', 'build', DOCS_DIR / '_build', DOCS_DIR / 'modules', Path('test-reports')]

PYTHON_VERSIONS = [
    '3.15',
    '3.14',
    '3.13',
    '3.12',
    '3.11',
    '3.10',
    '3.9',
    '3.8',
    'pypy3.11',
    'pypy3.10',
    'pypy3.9',
]
UNIT_TESTS = TEST_DIR / 'unit'
INTEGRATION_TESTS = TEST_DIR / 'integration'
COMPAT_TESTS = TEST_DIR / 'compat'
ALL_TESTS = [UNIT_TESTS, INTEGRATION_TESTS, COMPAT_TESTS]
STRESS_TEST_MULTIPLIER = 10
DEFAULT_COVERAGE_FORMATS = ['html', 'term']
# Run tests in parallel, grouped by test module
XDIST_ARGS = ['--numprocesses=auto', '--dist=loadfile']

IS_PYPY = platform.python_implementation() == 'PyPy'


def install_deps(session):
    """Install project and test dependencies into a test-specific virtualenv using uv"""
    session.env['UV_PROJECT_ENVIRONMENT'] = session.virtualenv.location
    session.run_install(
        'uv',
        'sync',
        '--frozen',
        '--all-extras',
    )


@nox.session(python=PYTHON_VERSIONS, venv_backend='uv')
def test(session):
    """Run tests in a separate virtualenv per python version"""
    test_paths = session.posargs or ALL_TESTS
    install_deps(session)
    session.run('pytest', '-rsxX', *XDIST_ARGS, *test_paths)


@nox.session(python=False, name='test-current')
def test_current(session):
    """Run tests using the current virtualenv"""
    test_paths = session.posargs or ALL_TESTS
    session.run('python', '-m', 'pytest', '-rsxX', *XDIST_ARGS, *test_paths)


@nox.session(python=False)
def clean(session):
    """Clean up temporary build + documentation files"""
    for dir in CLEAN_DIRS:
        print(f'Removing {dir}')
        rmtree(dir, ignore_errors=True)


@nox.session(python=False, name='cov')
def coverage(session):
    """Run tests and generate coverage report"""
    cmd = ['pytest', *ALL_TESTS, '-rsxX']

    # Exclude concurrency tests to run separately without xdist
    cmd += ['-k', 'not concurrency']
    if not IS_PYPY:
        cmd += XDIST_ARGS

    # Add coverage formats
    cmd += ['--cov']
    cov_formats = session.posargs or DEFAULT_COVERAGE_FORMATS
    cmd += [f'--cov-report={f}' for f in cov_formats]

    # Add verbose flag, if set by environment
    if getenv('PYTEST_VERBOSE'):
        cmd += ['--verbose']
    session.run(*cmd)


@nox.session(python=False, name='stress')
def stress_test(session):
    """Run concurrency tests with a higher stress test multiplier"""
    multiplier = session.posargs[0] if session.posargs else STRESS_TEST_MULTIPLIER
    cmd = ['pytest', INTEGRATION_TESTS, '-rs', '-k', 'concurrency']
    session.run(
        *cmd,
        env={'STRESS_TEST_MULTIPLIER': str(multiplier)},
    )


@nox.session(python=False)
def docs(session):
    """Build Sphinx documentation"""
    session.run('sphinx-build', 'docs', DOC_BUILD_DIR, '-j', 'auto')


@nox.session(python=False)
def linkcheck(session):
    """Check documentation for dead links"""
    session.run('sphinx-build', 'docs', DOC_BUILD_DIR, '-b', 'linkcheck')


@nox.session(python=False)
def livedocs(session):
    """Auto-build docs with live reload in browser.
    Add `--open` to also open the browser after starting.
    """
    cmd = ['sphinx-autobuild', 'docs', 'docs/_build/html']
    cmd += ['-a']
    cmd += ['--host', '0.0.0.0']
    cmd += ['--port', str(LIVE_DOCS_PORT), '-j', 'auto']
    for pattern in LIVE_DOCS_WATCH:
        cmd += ['--watch', pattern]
    for pattern in LIVE_DOCS_IGNORE:
        cmd += ['--ignore', pattern]
    if session.posargs == ['open']:
        cmd.append('--open-browser')

    clean(session)
    session.run(*cmd)


@nox.session(python=False)
def lint(session):
    """Run linters and code formatters via pre-commit"""
    session.run('prek', 'run', '--all-files')



#!/usr/bin/env python3
"""
An example of using the [time-machine](https://github.com/adamchainz/time-machine) library for backtesting,
e.g., testing with cached responses that were available at an arbitrary time in the past.
"""
from datetime import UTC, datetime

import requests
import time_machine

from requests_cache import CachedSession, set_response_defaults


class BacktestCachedSession(CachedSession):
    def request(self, method: str, url: str, **kwargs):
        response = super().request(method, url, **kwargs)

        # Response was cached after the (simulated) current time, so ignore it and send a new request
        if response.created_at and response.created_at > datetime.now(UTC):
            new_response = requests.request(method, url, **kwargs)
            return set_response_defaults(new_response)
        else:
            return response


def demo():
    session = BacktestCachedSession()
    response = session.get('https://httpbin.org/get')
    response = session.get('https://httpbin.org/get')
    assert response.from_cache is True

    # Response was not cached yet at this point, so we should get a fresh one
    with time_machine.travel(datetime(2020, 1, 1)):
        response = session.get('https://httpbin.org/get')
        assert response.from_cache is False


if __name__ == '__main__':
    demo()



from unittest.mock import patch

import requests
from requests.sessions import Session as OriginalSession

import requests_cache
from requests_cache import CachedSession
from requests_cache.backends import BaseCache, SQLiteCache
from tests.conftest import CACHE_NAME


def test_install_uninstall():
    for _ in range(2):
        requests_cache.install_cache(CACHE_NAME, use_temp=True)
        assert isinstance(requests.Session(), CachedSession)
        assert isinstance(requests.sessions.Session(), CachedSession)
        requests_cache.uninstall_cache()
        assert not isinstance(requests.Session(), CachedSession)
        assert not isinstance(requests.sessions.Session(), CachedSession)


def test_session_settings():
    requests_cache.install_cache(
        CACHE_NAME,
        use_temp=True,
        expire_after=360,
        cache_control=True,
        allowable_codes=(200, 301),
    )

    patched_session = requests.Session()
    assert patched_session.cache.cache_name == CACHE_NAME
    assert patched_session.settings.expire_after == 360
    assert patched_session.settings.cache_control is True
    assert patched_session.settings.allowable_codes == (200, 301)
    assert patched_session.settings.autoclose is False

    requests_cache.uninstall_cache()


def test_session_is_a_class_with_original_attributes(installed_session):
    assert isinstance(requests.Session, type)
    for attribute in dir(OriginalSession):
        assert hasattr(requests.Session, attribute)
    assert isinstance(requests.Session(), CachedSession)


def test_inheritance_after_monkey_patch(installed_session):
    class FooSession(requests.Session):
        __attrs__ = requests.Session.__attrs__ + ['new_one']

        def __init__(self, param):
            self.param = param
            super(FooSession, self).__init__()

    s = FooSession(1)
    assert isinstance(s, CachedSession)
    assert s.param == 1
    assert 'new_one' in s.__attrs__


@patch.object(SQLiteCache, 'clear')
def test_clear(mock_clear, installed_session):
    requests_cache.clear()
    mock_clear.assert_called()


@patch.object(SQLiteCache, 'clear')
def test_clear__not_installed(mock_clear):
    """If clear is called without a cache installed, it should just fail silently"""
    requests_cache.clear()
    mock_clear.assert_not_called()


@patch.object(OriginalSession, 'request')
@patch.object(CachedSession, 'request')
def test_disabled(cached_request, original_request, installed_session):
    with requests_cache.disabled():
        for _ in range(3):
            requests.get('some_url')
    assert cached_request.call_count == 0
    assert original_request.call_count == 3


@patch.object(OriginalSession, 'request')
@patch.object(CachedSession, 'request')
def test_enabled(cached_request, original_request, tempfile_path):
    with requests_cache.enabled(tempfile_path):
        for _ in range(3):
            requests.get('some_url')
    assert cached_request.call_count == 3
    assert original_request.call_count == 0


def test_is_installed():
    assert requests_cache.is_installed() is False
    requests_cache.install_cache(CACHE_NAME, use_temp=True)
    assert requests_cache.is_installed() is True
    requests_cache.uninstall_cache()
    assert requests_cache.is_installed() is False


@patch.object(BaseCache, 'delete')
def test_delete__expired_responses(mock_delete):
    requests_cache.install_cache(backend='memory', expire_after=360)
    requests_cache.delete(expired=True)
    assert mock_delete.called is True
    requests_cache.uninstall_cache()


@patch.object(BaseCache, 'delete')
def test_delete__cache_not_installed(mock_delete):
    requests_cache.delete(expired=True)
    assert mock_delete.called is False



from datetime import timedelta
from unittest.mock import patch

import pytest
from requests import PreparedRequest, Request
from requests.cookies import RequestsCookieJar

from requests_cache.cache_keys import create_key
from requests_cache.models import CachedRequest, CachedResponse
from requests_cache.policy import EXPIRE_IMMEDIATELY, CacheActions, CacheSettings, utcnow
from tests.conftest import ETAG, HTTPDATE_STR, LAST_MODIFIED, MOCKED_URL, get_mock_response

IGNORED_DIRECTIVES = [
    'no-transform',
    'private',
    'proxy-revalidate',
    'public',
    's-maxage=<seconds>',
]
BASIC_REQUEST = Request(method='GET', url='https://site.com/img.jpg', headers={})
EXPIRED_RESPONSE = CachedResponse(expires=utcnow() - timedelta(1))


@pytest.mark.parametrize(
    'request_expire_after, url_expire_after, expected_expiration',
    [
        (2, 3, 2),
        (None, 3, 3),
        (2, None, 2),
        (None, None, 1),
    ],
)
@patch('requests_cache.policy.actions.get_url_expiration')
def test_init(
    get_url_expiration,
    request_expire_after,
    url_expire_after,
    expected_expiration,
):
    """Test precedence with various combinations or per-request, per-session, per-URL, and
    Cache-Control expiration
    """
    request = PreparedRequest()
    request.url = 'https://img.site.com/base/img.jpg'
    if request_expire_after:
        request.headers = {'Cache-Control': f'max-age={request_expire_after}'}
    get_url_expiration.return_value = url_expire_after

    settings = CacheSettings(cache_control=True, expire_after=1)
    actions = CacheActions.from_request(cache_key='key', request=request, settings=settings)
    assert actions.expire_after == expected_expiration


@pytest.mark.parametrize(
    'headers, expected_expiration',
    [
        ({}, None),
        ({'Expires': HTTPDATE_STR}, None),  # Only valid for response headers
        ({'Cache-Control': 'max-age=60'}, 60),
        ({'Cache-Control': 'public, max-age=60'}, 60),
        ({'Cache-Control': b'public, max-age=60'}, 60),  # requests-oauthlib casts headers to bytes
        ({'Cache-Control': 'max-age=0'}, EXPIRE_IMMEDIATELY),
    ],
)
def test_init_from_headers(headers, expected_expiration):
    """Test with Cache-Control request headers"""
    settings = CacheSettings(cache_control=True)
    request = Request(method='GET', url=MOCKED_URL, headers=headers).prepare()
    actions = CacheActions.from_request('key', request, settings)

    assert actions.cache_key == 'key'
    if expected_expiration != EXPIRE_IMMEDIATELY:
        assert actions.expire_after == expected_expiration
        assert actions.skip_read is False
        assert actions.skip_write is False


def test_init_from_headers__no_store():
    """Test with Cache-Control request headers"""
    settings = CacheSettings(cache_control=True)
    request = Request(method='GET', url=MOCKED_URL, headers={'Cache-Control': 'no-store'}).prepare()
    actions = CacheActions.from_request('key', request, settings)

    assert actions.skip_read is True
    assert actions.skip_write is True


@pytest.mark.parametrize(
    'url, request_expire_after, expected_expiration',
    [
        ('https://img.site_1.com', None, timedelta(hours=12)),
        ('https://img.site_1.com', 60, 60),
        ('https://img.site.com/base/', None, 1),
        ('https://img.site.com/base/img.jpg', None, 1),
        ('http://site_2.com/resource_1', None, timedelta(hours=20)),
        ('ftp://site_2.com/resource_1/index.html', None, timedelta(hours=20)),
        ('http://site_2.com/resource_2/', None, timedelta(days=7)),
        ('http://site_2.com/static/', None, -1),
        ('http://site_2.com/static/img.jpg', None, -1),
        ('http://site_2.com', None, 1),
        ('http://site_2.com', 60, 60),
        ('https://some_other_site.com', None, 1),
        ('https://some_other_site.com', 60, 60),
    ],
)
def test_init_from_settings(url, request_expire_after, expected_expiration):
    """Test with per-session, per-request, and per-URL expiration"""
    settings = CacheSettings(
        expire_after=1,
        urls_expire_after={
            '*.site_1.com': timedelta(hours=12),
            'site_2.com/resource_1': timedelta(hours=20),
            'site_2.com/resource_2': timedelta(days=7),
            'site_2.com/static': -1,
        },
    )
    request = Request(method='GET', url=url)
    if request_expire_after:
        request.headers = {'Cache-Control': f'max-age={request_expire_after}'}

    actions = CacheActions.from_request('key', request.prepare(), settings)
    assert actions.expire_after == expected_expiration


@pytest.mark.parametrize(
    'headers, expire_after, expected_expiration, expected_skip_read',
    [
        ({'Cache-Control': 'max-age=60'}, 1, 60, False),
        ({}, 1, 1, False),
        ({}, 0, 0, False),
        ({'Cache-Control': 'max-age=60'}, 1, 60, False),
        ({'Cache-Control': 'max-age=0'}, 1, 0, False),
        ({'Cache-Control': 'no-store'}, 1, 1, True),
        ({'Cache-Control': 'no-cache'}, 1, 1, True),
    ],
)
def test_init_from_settings_and_headers(
    headers, expire_after, expected_expiration, expected_skip_read
):
    """Test behavior with both cache settings and request headers."""
    request = Request(method='GET', url=MOCKED_URL, headers=headers)
    settings = CacheSettings(expire_after=expire_after)
    actions = CacheActions.from_request('key', request, settings)

    assert actions.expire_after == expected_expiration
    assert actions.skip_read == expected_skip_read


def test_update_from_cached_response__new_request():
    actions = CacheActions.from_request('key', BASIC_REQUEST)
    actions.update_from_cached_response(None)
    assert actions.send_request is True


def test_update_from_cached_response__resend_request():
    actions = CacheActions.from_request('key', BASIC_REQUEST)

    actions.update_from_cached_response(EXPIRED_RESPONSE)
    assert actions.resend_request is True


@pytest.mark.parametrize(
    'response_headers, expected_validation_headers',
    [
        ({}, {}),
        ({'ETag': ETAG}, {'If-None-Match': ETAG}),
        ({'Last-Modified': LAST_MODIFIED}, {'If-Modified-Since': LAST_MODIFIED}),
        (
            {'ETag': ETAG, 'Last-Modified': LAST_MODIFIED},
            {'If-None-Match': ETAG, 'If-Modified-Since': LAST_MODIFIED},
        ),
    ],
)
def test_update_from_cached_response__revalidate(response_headers, expected_validation_headers):
    """Conditional request headers should be added if the cached response is expired"""
    actions = CacheActions.from_request('key', BASIC_REQUEST)
    cached_response = CachedResponse(headers=response_headers, expires=utcnow() - timedelta(1))

    actions.update_from_cached_response(cached_response)
    assert actions.send_request is False
    assert actions.resend_request is True
    assert actions._validation_headers == expected_validation_headers


@pytest.mark.parametrize(
    'response_headers',
    [
        {'Cache-Control': 'no-cache'},
        {'Cache-Control': 'max-age=0,must-revalidate'},
    ],
)
@pytest.mark.parametrize('cache_control', [True, False])
def test_update_from_cached_response__force_revalidate(cache_control, response_headers):
    """Conditional request headers should be added if requested by response headers, even if the
    response is not expired
    """
    actions = CacheActions.from_request(
        'key',
        request=Request(url='https://img.site.com/base/img.jpg', headers={}),
        settings=CacheSettings(cache_control=cache_control),
    )
    cached_response = CachedResponse(headers={'ETag': ETAG, **response_headers}, expires=None)

    actions.update_from_cached_response(cached_response)

    # cache_control=False overrides revalidation in this case
    if cache_control is False:
        assert actions.resend_request is False
        assert not actions._validation_headers
    else:
        assert actions.resend_request is True
        assert actions._validation_headers == {'If-None-Match': ETAG}


def test_update_from_cached_response__no_revalidation():
    """Conditional request headers should NOT be added if the cached response is not expired and
    revalidation is otherwise not requested"""
    actions = CacheActions.from_request('key', BASIC_REQUEST)
    cached_response = CachedResponse(
        headers={'ETag': ETAG, 'Last-Modified': LAST_MODIFIED}, expires=None
    )

    actions.update_from_cached_response(cached_response)
    assert actions._validation_headers == {}


def test_update_from_cached_response__504():
    settings = CacheSettings(only_if_cached=True)
    actions = CacheActions.from_request('key', BASIC_REQUEST, settings=settings)
    actions.update_from_cached_response(EXPIRED_RESPONSE)
    assert actions.error_504 is True


def test_update_from_cached_response__stale_if_error():
    settings = CacheSettings(only_if_cached=True, stale_if_error=True)
    actions = CacheActions.from_request('key', BASIC_REQUEST, settings=settings)
    actions.update_from_cached_response(EXPIRED_RESPONSE)
    assert actions.error_504 is False and actions.resend_request is False


def test_update_from_cached_response__stale_while_revalidate():
    settings = CacheSettings(only_if_cached=True, stale_while_revalidate=True)
    actions = CacheActions.from_request('key', BASIC_REQUEST, settings=settings)
    actions.update_from_cached_response(EXPIRED_RESPONSE)
    assert actions.resend_async is True


@pytest.mark.parametrize(
    'vary, cached_headers, new_headers, expected_match',
    [
        ({}, {}, {}, True),
        ({'Vary': 'Accept'}, {'Accept': 'application/json'}, {'Accept': 'application/json'}, True),
        ({'Vary': 'Accept'}, {'Accept': 'application/json'}, {}, False),
        (
            {'Vary': 'Accept'},
            {'Accept': 'application/json'},
            {'Accept': 'application/json', 'Accept-Language': 'en'},
            True,
        ),
        (
            {'Vary': 'Accept-Encoding'},
            {'Accept': 'application/json'},
            {'Accept': 'text/html'},
            True,
        ),
        ({'Vary': 'Accept'}, {'Accept': 'application/json'}, {'Accept': 'text/html'}, False),
        (
            {'Vary': 'Accept-Encoding'},
            {'Accept-Encoding': 'gzip,deflate'},
            {'Accept-Encoding': 'gzip,deflate'},
            True,
        ),
        # Only basic header normalization is done in create_key() (whitespace, case, order)
        (
            {'Vary': 'Accept-Encoding'},
            {'Accept-Encoding': 'gzip,deflate'},
            {'Accept-Encoding': 'dEfLaTe,  GZIP, '},
            True,
        ),
        (
            {'Vary': 'Accept-Encoding'},
            {'Accept-Encoding': 'gzip,deflate'},
            {'Accept-Encoding': 'gzip,br'},
            False,
        ),
        (
            {'Vary': 'Accept, Accept-Encoding'},
            {'Accept': 'application/json', 'Accept-Encoding': 'gzip,deflate'},
            {'Accept': 'application/json', 'Accept-Encoding': 'gzip,deflate'},
            True,
        ),
        (
            {'Vary': 'Accept, Accept-Encoding'},
            {'Accept': 'application/json', 'Accept-Encoding': 'gzip,deflate'},
            {'Accept': 'application/json', 'Accept-Encoding': 'br'},
            False,
        ),
        (
            {'Vary': 'Accept, Accept-Encoding'},
            {'Accept': 'application/json', 'Accept-Encoding': 'gzip,deflate'},
            {'Accept': 'text/html', 'Accept-Encoding': 'gzip,deflate'},
            False,
        ),
        (
            {'Vary': 'Accept, Accept-Encoding'},
            {'Accept': 'application/json', 'Accept-Encoding': 'gzip,deflate'},
            {'Accept-Encoding': 'gzip,deflate'},
            False,
        ),
        ({'Vary': '*'}, {}, {}, False),
        ({'Vary': '*'}, {'Accept': 'application/json'}, {'Accept': 'application/json'}, False),
    ],
)
def test_update_from_cached_response__vary_headers(
    vary, cached_headers, new_headers, expected_match
):
    cached_response = CachedResponse(
        headers=vary,
        request=Request(method='GET', url='https://site.com/img.jpg', headers=cached_headers),
    )
    request = Request(method='GET', url='https://site.com/img.jpg', headers=new_headers)
    actions = CacheActions.from_request('key', request)
    actions.update_from_cached_response(cached_response, create_key=create_key)

    # If the headers don't match wrt. Vary, expect a new request to be sent (cache miss)
    assert actions.send_request is not expected_match


@pytest.mark.parametrize(
    'vary, cached_headers, new_headers, expect_vary_key_set',
    [
        # Vary mismatch: secondary key should be set (and differ from base key)
        ({'Vary': 'Accept'}, {'Accept': 'application/json'}, {'Accept': 'text/html'}, True),
        # Vary match: no secondary key needed
        ({'Vary': 'Accept'}, {'Accept': 'application/json'}, {'Accept': 'application/json'}, False),
        # Vary: * always misses, but no secondary lookup is possible
        ({'Vary': '*'}, {'Accept': 'application/json'}, {'Accept': 'application/json'}, False),
        # No Vary header: no secondary key
        ({}, {}, {}, False),
        # Vary: Cookie with matching cookies: no secondary key (cookies match, so cache hit)
        ({'Vary': 'Cookie'}, {}, {}, False),
    ],
)
def test_vary_cache_key(vary, cached_headers, new_headers, expect_vary_key_set):
    cached_response = CachedResponse(
        headers=vary,
        request=Request(method='GET', url='https://site.com/img.jpg', headers=cached_headers),
    )
    request = Request(method='GET', url='https://site.com/img.jpg', headers=new_headers)
    actions = CacheActions.from_request('key', request)
    actions.update_from_cached_response(cached_response, create_key=create_key)

    if expect_vary_key_set:
        assert actions.vary_cache_key is not None
        assert actions.vary_cache_key != actions.cache_key
    else:
        assert actions.vary_cache_key is None


def test_vary_cache_key__merges_user_match_headers():
    """vary_cache_key should incorporate user match_headers along with Vary headers"""
    cached_response = CachedResponse(
        headers={'Vary': 'Accept'},
        request=Request(
            method='GET',
            url='https://site.com/img.jpg',
            headers={'Accept': 'application/json', 'Authorization': 'Bearer token1'},
        ),
    )
    request = Request(
        method='GET',
        url='https://site.com/img.jpg',
        headers={'Accept': 'text/html', 'Authorization': 'Bearer token2'},
    )
    settings = CacheSettings(match_headers=['Authorization'])
    actions = CacheActions.from_request('key', request, settings=settings)
    actions.update_from_cached_response(cached_response, create_key=create_key)

    assert actions.vary_cache_key is not None

    # The secondary key should differ from a key built with only Vary headers
    request_2 = Request(
        method='GET',
        url='https://site.com/img.jpg',
        headers={'Accept': 'text/html', 'Authorization': 'Bearer token_different'},
    )
    actions_2 = CacheActions.from_request('key', request_2, settings=settings)
    actions_2.update_from_cached_response(cached_response, create_key=create_key)
    assert actions_2.vary_cache_key != actions.vary_cache_key


def test_vary_cache_key__match_all():
    """vary_cache_key should be set when match_headers=True (match all headers)"""
    cached_response = CachedResponse(
        headers={'Vary': 'Accept'},
        request=Request(
            method='GET',
            url='https://site.com/img.jpg',
            headers={'Accept': 'application/json'},
        ),
    )
    request = Request(
        method='GET',
        url='https://site.com/img.jpg',
        headers={'Accept': 'text/html'},
    )
    settings = CacheSettings(match_headers=True)
    actions = CacheActions.from_request('key', request, settings=settings)
    actions.update_from_cached_response(cached_response, create_key=create_key)

    assert actions.vary_cache_key is not None
    assert actions.vary_cache_key != actions.cache_key


@pytest.mark.parametrize(
    'cached_cookies, new_cookies, expected_match',
    [
        ({'session': 'abc123'}, {'session': 'abc123'}, True),
        ({'session': 'abc123'}, {'session': 'xyz789'}, False),
        ({'session': 'abc', 'user': 'bob'}, {'session': 'abc', 'user': 'bob'}, True),
        ({'user': 'bob', 'session': 'abc'}, {'session': 'abc', 'user': 'bob'}, True),
        ({'session': 'abc', 'user': 'bob'}, {'session': 'abc', 'user': 'alice'}, False),
        ({}, {}, True),
        ({'session': 'abc123'}, {}, False),
        ({}, {'session': 'abc123'}, False),
    ],
)
def test_update_from_cached_response__vary_cookie(cached_cookies, new_cookies, expected_match):
    cached_cookiejar = RequestsCookieJar()
    cached_cookiejar.update(cached_cookies)
    new_cookiejar = RequestsCookieJar()
    new_cookiejar.update(new_cookies)

    # Cached response with cookies
    cached_request = CachedRequest(
        method='GET',
        url='https://site.com/page',
        cookies=cached_cookiejar,
    )
    cached_response = CachedResponse(
        headers={'Vary': 'Cookie'},
        request=cached_request,
    )

    # New request with cookies
    new_request = PreparedRequest()
    new_request.prepare(method='GET', url='https://site.com/page', cookies=new_cookiejar)

    actions = CacheActions.from_request('key', new_request)
    actions.update_from_cached_response(cached_response, create_key=create_key)

    # If cookies don't match, expect a new request (cache miss)
    assert actions.send_request is not expected_match


def test_update_from_cached_response__vary_cookie_and_headers():
    cached_cookiejar = RequestsCookieJar()
    cached_cookiejar.set('session', 'abc123')
    new_cookiejar = RequestsCookieJar()
    new_cookiejar.set('session', 'abc123')

    # Cached response with Vary: Cookie, Accept
    cached_request = CachedRequest(
        method='GET',
        url='https://site.com/page',
        headers={'Accept': 'application/json'},
        cookies=cached_cookiejar,
    )
    cached_response = CachedResponse(
        headers={'Vary': 'Cookie, Accept'},
        request=cached_request,
    )

    # New request with matching cookies but different Accept header
    new_request = PreparedRequest()
    new_request.prepare(
        method='GET',
        url='https://site.com/page',
        headers={'Accept': 'text/html'},
        cookies=new_cookiejar,
    )

    actions = CacheActions.from_request('key', new_request)
    actions.update_from_cached_response(cached_response, create_key=create_key)

    # Should be a cache miss because Accept header differs
    assert actions.send_request is True
    # Cookies match, so secondary key is set for the differing Accept header
    assert actions.vary_cache_key is not None


@pytest.mark.parametrize('max_stale, usable', [(5, False), (15, True)])
def test_is_usable__max_stale(max_stale, usable):
    """For a response that expired 10 seconds ago, it may be either accepted or rejected based on
    max-stale
    """
    request = Request(
        url='https://img.site.com/base/img.jpg',
        headers={'Cache-Control': f'max-stale={max_stale}'},
    )
    actions = CacheActions.from_request('key', request)
    cached_response = CachedResponse(expires=utcnow() - timedelta(seconds=10))
    assert actions.is_usable(cached_response) is usable


@pytest.mark.parametrize('min_fresh, usable', [(5, True), (15, False)])
def test_is_usable__min_fresh(min_fresh, usable):
    """For a response that expires in 10 seconds, it may be either accepted or rejected based on
    min-fresh
    """
    request = Request(
        url='https://img.site.com/base/img.jpg',
        headers={'Cache-Control': f'min-fresh={min_fresh}'},
    )
    actions = CacheActions.from_request('key', request)
    cached_response = CachedResponse(expires=utcnow() + timedelta(seconds=10))
    assert actions.is_usable(cached_response) is usable


@pytest.mark.parametrize(
    'stale_if_error, error, usable',
    [
        (5, True, False),
        (15, True, True),
        (15, False, False),
    ],
)
def test_is_usable__stale_if_error(stale_if_error, error, usable):
    """For a response that expired 10 seconds ago, if an error occurred while refreshing, it may be
    either accepted or rejected based on stale-if-error
    """
    request = Request(
        url='https://img.site.com/base/img.jpg',
        headers={'Cache-Control': f'stale-if-error={stale_if_error}'},
    )
    actions = CacheActions.from_request('key', request)
    cached_response = CachedResponse(expires=utcnow() - timedelta(seconds=10))
    assert actions.is_usable(cached_response, error=error) is usable


@pytest.mark.parametrize(
    'stale_while_revalidate, usable',
    [
        (5, False),
        (15, True),
    ],
)
def test_is_usable__stale_while_revalidate(stale_while_revalidate, usable):
    """For a response that expired 10 seconds ago, if an error occurred while refreshing, it may be
    either accepted or rejected based on stale-while-revalidate
    """
    request = Request(
        url='https://img.site.com/base/img.jpg',
        headers={'Cache-Control': f'stale-while-revalidate={stale_while_revalidate}'},
    )
    actions = CacheActions.from_request('key', request)
    cached_response = CachedResponse(expires=utcnow() - timedelta(seconds=10))
    assert actions.is_usable(cached_response=cached_response) is usable


@pytest.mark.parametrize(
    'headers, expected_expiration',
    [
        ({}, None),
        ({'Cache-Control': 'no-cache'}, None),  # Forces revalidation, but no effect on expiration
        ({'Cache-Control': 'max-age=0'}, 0),
        ({'Cache-Control': 'max-age=60'}, 60),
        ({'Cache-Control': 'public, max-age=60'}, 60),
        ({'Cache-Control': 'max-age=0'}, 0),
        ({'Cache-Control': 'immutable'}, -1),
        ({'Cache-Control': 'immutable, max-age=60'}, -1),  # Immutable should take precedence
        ({'Expires': HTTPDATE_STR}, HTTPDATE_STR),
        ({'Expires': HTTPDATE_STR, 'Cache-Control': 'max-age=60'}, 60),
    ],
)
def test_update_from_response(headers, expected_expiration):
    """Test with Cache-Control response headers"""
    actions = CacheActions.from_request('key', BASIC_REQUEST, CacheSettings(cache_control=True))
    actions.update_from_response(get_mock_response(headers=headers))

    assert actions.expire_after == expected_expiration
    assert actions.skip_write is (expected_expiration == EXPIRE_IMMEDIATELY)


def test_update_from_response__no_store():
    actions = CacheActions.from_request('key', BASIC_REQUEST, CacheSettings(cache_control=True))
    actions.update_from_response(get_mock_response(headers={'Cache-Control': 'no-store'}))
    assert actions.skip_write is True


def test_update_from_response__ignored():
    actions = CacheActions.from_request('key', BASIC_REQUEST, CacheSettings(cache_control=False))
    actions.update_from_response(get_mock_response(headers={'Cache-Control': 'max-age=5'}))
    assert actions.expire_after is None


@pytest.mark.parametrize('validator_headers', [{'ETag': ETAG}, {'Last-Modified': LAST_MODIFIED}])
@pytest.mark.parametrize('cache_headers', [{'Cache-Control': 'max-age=0'}, {'Expires': '0'}])
@patch('requests_cache.expiration.utcnow')
def test_update_from_response__revalidate(mock_utcnow, cache_headers, validator_headers):
    """If expiration is 0 and there's a validator, the response should be cached, but with immediate
    expiration
    """
    actions = CacheActions.from_request('key', BASIC_REQUEST, CacheSettings(cache_control=True))
    response = get_mock_response(headers={**cache_headers, **validator_headers})
    actions.update_from_response(response)

    assert actions.expires == mock_utcnow()
    assert actions.skip_write is False


@pytest.mark.parametrize(
    'read_only, headers_changed, expected_skip_write',
    [
        (False, False, True),  # Nothing changed: skip_write should be True (optimization)
        (False, True, False),  # Headers changed, not read_only: should write
        (True, False, True),  # read_only: should remain True even if nothing changed
        (True, True, True),  # read_only + headers changed: skip_write must remain True
    ],
)
def test_update_revalidated_response__skip_write(read_only, headers_changed, expected_skip_write):
    """skip_write should be True when read_only=True or when nothing changed,
    and False when headers/expiration changed and read_only=False
    """
    actions = CacheActions.from_request(
        'key', BASIC_REQUEST, CacheSettings(cache_control=True, read_only=read_only)
    )
    cached_response = get_mock_response(headers={'ETag': ETAG})
    cached_response.expires = None
    new_headers = {'ETag': ETAG, 'X-Custom-Header': 'new'} if headers_changed else {'ETag': ETAG}
    new_response = get_mock_response(status_code=304, headers=new_headers)
    actions.update_revalidated_response(new_response, cached_response)
    assert actions.skip_write is expected_skip_write


def test_update_revalidated_response__transfer_encoding():
    """When updating response headers after revalidating a cached response,
    don't set both `Content-Length` and `Transfer-Encoding`
    """
    actions = CacheActions.from_request('key', BASIC_REQUEST, CacheSettings(cache_control=True))
    cached_response = get_mock_response(headers={'ETag': ETAG, 'Transfer-Encoding': 'chunked'})
    new_response = get_mock_response(status_code=304, headers={'ETag': ETAG, 'Content-Length': 0})
    cached_response = actions.update_revalidated_response(new_response, cached_response)
    assert cached_response.headers['Transfer-Encoding'] == 'chunked'
    assert 'Content-Length' not in cached_response.headers


@pytest.mark.parametrize('directive', IGNORED_DIRECTIVES)
def test_ignored_headers(directive):
    """Ensure that currently unimplemented Cache-Control headers do not affect behavior"""
    request = Request(
        method='GET', url='https://img.site.com/base/img.jpg', headers={'Cache-Control': directive}
    ).prepare()
    settings = CacheSettings(expire_after=1, cache_control=True)
    actions = CacheActions.from_request('key', request, settings)

    assert actions.expire_after == 1
    assert actions.skip_read is False
    assert actions.skip_write is False
